import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    EXPLICIT_NULL,
    collection_create_and_update_params,
    globus_command,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import (
    CollectionDocument,
    GoogleCloudStorageCollectionPolicies,
)
from gcs_cli.services.gcs import get_client


@globus_command("create", short_help="Create a new (mapped) Collection")
@click.argument("STORAGE_GATEWAY_ID")
@click.argument("BASE_PATH")
@collection_create_and_update_params(create=True)
@click.option(
    "--identity-id",
    help=(
        "Globus Auth identity to who acts as the owner of this Collection. This "
        "identity is an administrator on the Endpoint"
    ),
)
@click.option(
    "--google-project-id",
    help=(
        "For Google Cloud Storage backed Collections only. The Google Cloud Platform "
        "project ID which is used by this Collection"
    ),
)
@requires_login
def collection_create(
    storage_gateway_id,
    base_path,
    identity_id,
    google_project_id,
    sharing_restrict_paths=None,
    user_message=None,
    user_message_link=None,
    **kwargs
):
    """
    Create a new Mapped Collection, rooted on some given path within an
    existing Storage Gateway.

    The STORAGE_GATEWAY_ID is the Storage Gateway which will host the Collection.
    """
    if identity_id is not None:
        kwargs["identity_id"] = identity_id
    if google_project_id is not None:
        kwargs["policies"] = GoogleCloudStorageCollectionPolicies(
            project=google_project_id
        )
    if sharing_restrict_paths is not None:
        if sharing_restrict_paths == EXPLICIT_NULL:
            kwargs["sharing_restrict_paths"] = None
        else:
            kwargs["sharing_restrict_paths"] = sharing_restrict_paths

    if user_message is not None:
        if user_message == EXPLICIT_NULL:
            kwargs["user_message"] = None
        else:
            kwargs["user_message"] = user_message

    if user_message_link is not None:
        if user_message_link == EXPLICIT_NULL:
            kwargs["user_message_link"] = None
        else:
            kwargs["user_message_link"] = user_message_link

    # These are added in GCS 5.4.4, so if neither --enable-https or
    # --disable-https are set, then we'll not try to set the value for
    # backward compatibility
    if kwargs.get("enable_https") and kwargs.get("disable_https"):
        raise click.exceptions.BadOptionUsage(
            "Only one of --enable-https and --disable-https may be used"
        )
    if kwargs.get("enable_https") is False:
        kwargs.pop("enable_https")

    if kwargs.get("disable_https"):
        kwargs["enable_https"] = False

    kwargs.pop("disable_https", None)

    collection_doc = CollectionDocument(
        storage_gateway_id=storage_gateway_id,
        collection_type="mapped",
        collection_base_path=base_path,
        **kwargs
    )
    client = get_client()
    res = client.create_collection(collection_doc)

    # output
    formatted_print(
        res, fields=[("Collection ID", "id")], text_format=FORMAT_TEXT_RECORD
    )
