import click

from gcs_cli.helpers import record_print_result_message, requires_login
from gcs_cli.parsing import NetworkPortRange, globus_command, node_id_arg
from gcs_cli.sdk_client import GCSAPIError
from gcs_cli.services.gcs import get_client


@globus_command("update", short_help="Update a Node's config, including enable/disable")
@node_id_arg
@click.option(
    "--enable/--disable",
    default=None,
    help="Set the node state to 'active' or 'inactive'",
)
@click.option(
    "--ip-address",
    "ip_addresses",
    multiple=True,
    help="IP address for the Node. Use this option multiple times to set multiple IPs",
)
@click.option(
    "--incoming-port-range",
    help="Allowed port range for incoming TCP data connections",
    # NOTE: hidden for now because the gridftp config to use the port range won't be
    # updated when this happens. Un-hide when GCS Manager can write the nodes' gridftp
    # config
    hidden=True,
    type=NetworkPortRange(),
)
@click.option(
    "--outgoing-port-range",
    help="Port range used as the source for outgoing TCP data connections",
    # NOTE: hidden for now because the gridftp config to use the port range won't be
    # updated when this happens. Un-hide when GCS Manager can write the nodes' gridftp
    # config
    hidden=True,
    type=NetworkPortRange(),
)
@requires_login
def node_update(
    node_id, enable, ip_addresses, incoming_port_range, outgoing_port_range
):
    """
    Update a Node on the current Endpoint.
    """
    data = {"id": str(node_id)}
    if enable is not None:
        data["status"] = "active" if enable else "inactive"
    if ip_addresses:
        data["ip_addresses"] = list(ip_addresses)
    if incoming_port_range:
        data["incoming_port_range"] = list(incoming_port_range)
    if outgoing_port_range:
        data["outgoing_port_range"] = list(outgoing_port_range)
    client = get_client()
    try:
        res = client.update_node(node_id, data)
    except GCSAPIError as e:
        if enable and e.http_status == 503:
            click.echo(
                (
                    "Node enable failed with an error which may indicate that you are "
                    "trying to enable the current host after it has been disabled. "
                    "This is not possible."
                    "\n\nYou should re-run"
                    "\n  globus-connect-server node setup\n"
                    "instead."
                ),
                err=True,
            )
        raise
    record_print_result_message(res)
