import click

from gcs_cli.parsing.types import JSONStringOrFile, nullable_multi_callback


def posix_options(create=False):
    # see shared_options.storage_gateway_create_and_update_params for an explanation of
    # the need for these two variants
    multi_use_option_str = "Give this option multiple times"
    if not create:
        multi_use_option_str = "Give this option multiple times in a single command"

    def inner_decorator(f):
        # POSIX policies
        f = click.option(
            "--posix-group-allow",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "POSIX group allowed access to this Storage Gateway. "
                "{} to allow multiple groups. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)
        f = click.option(
            "--posix-group-deny",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "POSIX group denied access to this Storage Gateway. "
                "{} to deny multiple groups. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)
        return f

    return inner_decorator


def _s3_endpoint_option(f, name):
    return click.option(
        "--s3-endpoint", help="Region-specific URI of the {} API".format(name)
    )(f)


def _bucket_option(f, name):
    connector_text = ""
    if name == "S3":
        connector_text = (
            "This is the only way to make public S3 buckets "
            "visible to the root of an S3 Storage Gateway. "
        )
    return click.option(
        "--bucket",
        multiple=True,
        callback=nullable_multi_callback(),
        help=(
            "Bucket to include in the root of the Storage Gateway. "
            "Give this option multiple times to include multiple "
            "buckets. "
        )
        + connector_text
        + 'Set a value of "null" to clear this',
    )(f)


def _google_options(f, name):
    f = click.option(
        "--google-client-id",
        help=("Application Client Id registered with Google to access " + name),
    )(f)
    f = click.option(
        "--google-client-secret",
        help=(
            "Secret created by Google to access Google as "
            "the given --google-client-id"
        ),
    )(f)
    return f


def blackpearl_options(f):
    f = _s3_endpoint_option(f, "BlackPearl")
    f = click.option(
        "--blackpearl-access-id-file", help="Path to the BlackPearl user mapping file"
    )(f)
    return f


def box_options(f):
    def validate_box_settings(ctx, param, value):
        if value is None:
            return None
        if "enterpriseID" not in value:
            raise click.UsageError("--box-settings must contain enterpriseID")
        if "boxAppSettings" not in value:
            raise click.UsageError("--box-settings must contain boxAppSettings")
        return value

    return click.option(
        "--box-settings",
        callback=validate_box_settings,
        type=JSONStringOrFile(),
        help="The Box App Settings JSON data, as a string or a file",
    )(f)


def ceph_options(f):
    f = _s3_endpoint_option(f, "Ceph")
    f = _bucket_option(f, "Ceph")
    f = click.option(
        "--ceph-admin-key-id",
        help="Key ID of an admin key used to resolve Ceph usernames to credentials",
    )(f)
    f = click.option(
        "--ceph-admin-secret-key",
        help="Secret key corresponding to --ceph-admin-key-id",
    )(f)
    return f


def google_drive_options(f):
    f = _google_options(f, "Google Drive")
    f = click.option(
        "--google-drive-user-api-rate-quota",
        type=int,
        help="Negotiated rate quota for Queries per 100 seconds per use",
    )(f)
    return f


def google_cloud_storage_options(f):
    f = _bucket_option(f, "Google Cloud Storage")
    f = _google_options(f, "Google Cloud Storage")
    f = click.option(
        "--google-cloud-storage-project",
        multiple=True,
        callback=nullable_multi_callback(),
        help=(
            "Project this Storage Gateway is allowed to access. "
            "Give this option multiple times to allow multiple "
            "projects. If no projects are specified, project access is "
            "unrestricted. "
            'Set a value of "null" to clear this'
        ),
    )(f)
    return f


def s3_options(f):
    f = _s3_endpoint_option(f, "S3")
    f = _bucket_option(f, "S3")
    f = click.option(
        "--s3-user-credential/--s3-unauthenticated",
        default=None,
        help=(
            "Either require an S3 user credential to access this "
            "Storage Gateway or not. If unauthenticated, the Storage "
            "Gateway will only allow access to public S3 buckets"
        ),
    )(f)
    return f
