import click

from gcs_cli.parsing.types import (
    CommaDelimitedList,
    IdentityMappingType,
    JSONStringOrFile,
    StringOrNull,
    UrlOrNull,
    UUIDorDEFAULTorNull,
    UUIDorUsername,
    nullable_multi_callback,
)


def _makehelp(txt, object_name, *, create):
    if create:
        return txt[0].upper() + txt[1:] + " " + object_name
    else:
        return "New " + txt + " " + object_name


def _display_name(f, object_name, *, create):
    """Options for Storage Gateway, Collection, Endpoint"""
    if create:
        f = click.argument("display_name")(f)
    else:
        f = click.option("--display-name", help="New name for the " + object_name)(f)

    return f


def _public(f, object_name, *, default=None):
    return click.option(
        "--public/--private",
        "public",
        default=default,
        help="Set the {} to be public or private{}".format(
            object_name, "" if not default else " (defaults to public)"
        ),
    )(f)


def _common_searchable_options(f, object_name, create=True):
    def _help(txt):
        return _makehelp(txt, object_name, create=create)

    f = _public(f, object_name, default=(True if create else None))
    f = click.option(
        "--description", type=StringOrNull(), help=_help("description for the")
    )(f)
    f = click.option(
        "--info-link", type=StringOrNull(), help=_help("link for info about the")
    )(f)
    f = click.option(
        "--contact-info", type=StringOrNull(), help=_help("contact Info for the")
    )(f)
    f = click.option(
        "--contact-email", type=StringOrNull(), help=_help("contact email for the")
    )(f)
    f = click.option(
        "--organization", type=StringOrNull(), help=_help("organization for the")
    )(f)
    f = click.option(
        "--department", type=StringOrNull(), help=_help("department which operates the")
    )(f)
    f = click.option(
        "--keywords",
        type=CommaDelimitedList(),
        help=_help("comma separated list of keywords to help searches for the"),
    )(f)

    return f


def collection_create_and_update_params(create=True):
    """
    Collection of options consumed by GCS Collection update

    Usage:

    >>> @collection_create_and_update_params(create=False)
    >>> def command_func(**kwargs):
    >>>     ...
    """

    def inner_decorator(f):
        f = _display_name(f, "Collection", create=create)
        f = _common_searchable_options(f, "Collection", create=create)
        f = click.option(
            "--force-encryption/--no-force-encryption",
            "force_encryption",
            default=None,
            help=(
                "When set, all transfers to and from this collection are "
                "always encrypted"
            ),
        )(f)
        f = click.option(
            "--sharing-restrict-paths",
            type=JSONStringOrFile(null="null"),
            help="Path restrictions for sharing data on guest collections "
            "based on this collection. This option is only usable on Mapped "
            "Collections",
        )(f)
        f = click.option(
            "--allow-guest-collections/--no-allow-guest-collections",
            "allow_guest_collections",
            default=None,
            help=(
                "Allow Guest Collections to be created on this Collection. This option "
                "is only usable on Mapped Collections. If this option is disabled on a "
                "Mapped Collection which already has associated Guest Collections, "
                "those collections will no longer be accessible"
            ),
        )(f)
        f = click.option(
            "--domain-name",
            "domain_name",
            default=None,
            help=(
                "DNS host name for the collection (mapped "
                "collections only). This may be either a host name "
                "or a fully-qualified domain name, but if it is the latter "
                "it must be a subdomain of the endpoint's domain"
            ),
        )(f)
        f = click.option(
            "--default-directory",
            default=None,
            help=("Default directory when browsing the collection"),
        )(f)
        f = click.option(
            "--enable-https",
            is_flag=True,
            help=(
                "Explicitly enable HTTPS supprt (requires a managed endpoint "
                "with API v1.1.0)"
            ),
        )(f)

        f = click.option(
            "--disable-https",
            is_flag=True,
            help=(
                "Explicitly disable HTTPS supprt (requires a managed endpoint "
                "with API v1.1.0)"
            ),
        )(f)
        f = click.option(
            "--user-message",
            help=(
                "A message for clients to display to users when interacting "
                "with this collection"
            ),
            type=StringOrNull(),
        )(f)
        f = click.option(
            "--user-message-link",
            help=(
                "Link to additional messaging for clients to display to users "
                "when interacting with this endpoint, linked to an http or https URL "
                "with this collection"
            ),
            type=UrlOrNull(),
        )(f)

        return f

    return inner_decorator


def endpoint_update_params(f):
    """
    Collection of options consumed by GCS endpoint update
    Unlike Transfer create_or_update options in the Globus CLI, Update is the
    only supported operation here
    """

    f = _display_name(f, "Endpoint", create=False)
    f = _common_searchable_options(f, "Endpoint", create=False)
    f = click.option(
        "--subscription-id",
        type=UUIDorDEFAULTorNull(),
        default=None,
        help=(
            "Set the Endpoint as a managed Endpoint with the given "
            "subscription ID or DEFAULT. Pass '--subscription-id DEFAULT' "
            "if are subscription manager for only one subscription, or pass "
            "'--subscription-id null' to set an endpoint's subscription ID "
            "to null"
        ),
    )(f)
    f = click.option(
        "--network-use",
        default=None,
        type=click.Choice(["normal", "minimal", "aggressive", "custom"]),
        help=(
            "Set the Endpoint's network use level. If using custom, "
            "the Endpoint's max and preferred concurrency and "
            "parallelism must be set"
        ),
    )(f)
    f = click.option(
        "--max-concurrency",
        type=int,
        default=None,
        help=("Set the Endpoint's max concurrency; requires --network-use=custom"),
    )(f)
    f = click.option(
        "--preferred-concurrency",
        type=int,
        default=None,
        help=(
            "Set the Endpoint's preferred concurrency; " "requires --network-use=custom"
        ),
    )(f)
    f = click.option(
        "--max-parallelism",
        type=int,
        default=None,
        help=("Set the Endpoint's max parallelism; requires --network-use=custom"),
    )(f)
    f = click.option(
        "--preferred-parallelism",
        type=int,
        default=None,
        help=(
            "Set the Endpoint's preferred parallelism; " "requires --network-use=custom"
        ),
    )(f)

    return f


def storage_gateway_create_and_update_params(create=False):
    """
    Collection of options consumed by GCS storage_gateway create and update
    operations -- accepts toggle regarding create vs. update to handle
    options required for create but not update
    """

    # alter the description by create vs update
    # update specifies that multi-use options are "in a single command"
    # see https://github.com/globusonline/gcsv5/issues/444 for details on the
    # potential point of confusion -- users may expect a cumulative application with
    # 'update'
    multi_use_option_str = "Give this option multiple times"
    if not create:
        multi_use_option_str = "Give this option multiple times in a single command"

    def inner_decorator(f):
        f = _display_name(f, "Storage Gateway", create=create)
        # some arguments are required on create, but optional
        # or unavailable on update
        if create:
            f = click.option(
                "--high-assurance",
                "high_assurance",
                is_flag=True,
                help=(
                    "Flag indicating that High Assurance features are "
                    "required on this storage gateway."
                ),
            )(f)

        # Options available to any storage_gateway
        f = click.option(
            "--authentication-timeout-mins",
            help=(
                "Timeout (in minutes) during which a user is required to "
                "have authenticated in a session to access this storage "
                "gateway."
            ),
        )(f)
        f = click.option(
            "--domain",
            "allowed_domains",
            multiple=True,
            callback=nullable_multi_callback(None),
            help=(
                "Allowed domain. {} to allow ".format(multi_use_option_str)
                + "multiple domains. Users creating credentials or "
                "collections on this storage gateway must have an identity "
                "in one of these domains."
            ),
        )(f)
        f = click.option(
            "--restrict-paths",
            type=JSONStringOrFile(null="null"),
            help="Path restrictions for accessing data on collections "
            "created using this storage gateway",
        )(f)
        f = click.option(
            "--identity-mapping",
            "identity_mappings",
            multiple=True,
            # exploit the 'null' -> None JSON parsing conversion to feed the callback
            # '[None]' if the user passes 'null'
            callback=nullable_multi_callback(None),
            type=IdentityMappingType(null="null"),
            help=(
                "An Identity Mapping for use in this Storage Gateway. "
                "You can use JSON input to specify a complete mapping document, "
                "or, if you want to use an external command for mapping, "
                "use 'external:<your command>'. "
                "{} to set multiple mappings in ".format(multi_use_option_str)
                + "order of precedence. "
                'Set a value of "null" to clear this'
            ),
        )(f)
        f = click.option(
            "--user-allow",
            "users_allow",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "Connector-specific username for a user allowed access to this "
                "Storage Gateway. "
                "{} to allow multiple users. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)
        f = click.option(
            "--user-deny",
            "users_deny",
            multiple=True,
            callback=nullable_multi_callback(""),
            help=(
                "Connector-specific username for a user denied access to this "
                "Storage Gateway. "
                "{} to deny multiple users. ".format(multi_use_option_str)
                + 'Set a value of "" to clear this'
            ),
        )(f)

        return f

    return inner_decorator


def role_create_params(endpoint=False):
    """
    Shared arguments for endpoint role create and collection role create
    """

    def inner_decorator(f):
        f = click.argument("principal_name_or_id", type=UUIDorUsername())(f)
        f = click.option(
            "--principal-type",
            type=click.Choice(("identity", "group"), case_sensitive=False),
            help=(
                'Type of principal to create the role for. If "identity" then '
                "this role will be granted to a single Globus identity, "
                'if "group" then this role will be granted to all members of the '
                "specified Globus group."
            ),
            default="identity",
            show_default=True,
        )(f)
        if endpoint:
            role_choices = ("administrator", "activity_manager", "activity_monitor")
        else:
            role_choices = (
                "administrator",
                "access_manager",
                "activity_manager",
                "activity_monitor",
            )
        f = click.argument(
            "role", type=click.Choice(role_choices, case_sensitive=False)
        )(f)

        return f

    return inner_decorator


def collection_id_arg(f):
    """
    COLLECTION_ID argument for collection show update and delete
    """
    return click.argument("collection_id", metavar="COLLECTION_ID", type=click.UUID)(f)


def storage_gateway_id_arg(f):
    """
    STORAGE_GATEWAY_ID argument for storage-gateway show update and delete
    """
    return click.argument(
        "storage_gateway_id", metavar="STORAGE_GATEWAY_ID", type=click.UUID
    )(f)


def role_id_arg(f):
    """
    ROLE_ID argument for role show and delete
    """
    return click.argument("role_id", metavar="ROLE_ID", type=click.UUID)(f)


def node_id_arg(f):
    """
    NODE_ID argument for node show update and delete
    """
    return click.argument("node_id", metavar="NODE_ID", type=click.UUID)(f)


def include_private_policies_opt(f):
    """
    --include-private-policies option for storage-gateway list and show
    """
    return click.option(
        "--include-private-policies",
        is_flag=True,
        help=(
            "Include private policies. Requires administrator role on the endpoint. "
            "Some policy data may only be visible in `--format JSON` output"
        ),
    )(f)
