import json
import shlex
from urllib.parse import urlparse

import click


class _ExplicitNullClass(object):
    """
    Magic sentinel value used to disambiguate values which are being
    intentionally nulled from values which are `None` because no argument was
    provided
    """

    def __bool__(self):
        return False

    def __repr__(self):
        return "null"


EXPLICIT_NULL = _ExplicitNullClass()


def nullable_multi_callback(null="null"):
    """
    A callback which converts multiple=True options as follows:
    - empty results, [] => None
    - [<null>,] => []
    - anything else => passthrough

    This makes the null value explicit, and not setting it results in omission, not
    clearing.

    The null value used here is tunable. If set to a non-string value when using a
    string type, like `None`, it means that "there is no null value" because
    there is no way to pass `[]`

    Note that this will see values after the type conversion has happened.
    """

    def callback(ctx, param, value):
        if value is None or len(value) == 0:
            return None
        if len(value) == 1 and value[0] == null:
            return []
        return value

    return callback


class UUIDorDEFAULTorNull(click.ParamType):
    def get_metavar(self, param):
        return "[UUID|DEFAULT|null]"

    def convert(self, value, param, ctx):
        if value is not None and value.lower() == "null":
            return EXPLICIT_NULL
        elif value is not None and value.lower() == "default":
            return "DEFAULT"
        else:
            return click.UUID.convert(value, param, ctx)


class UUIDorUsername(click.ParamType):
    def get_metavar(self, param):
        return "[UUID|USERNAME]"

    def convert(self, value, param, ctx):
        if value is not None and "@" in value:
            return value
        else:
            return click.UUID.convert(value, param, ctx)


class StringOrNull(click.ParamType):
    """
    Very similar to a basic string type, but one in which the empty string will
    be converted into an EXPLICIT_NULL
    """

    def get_metavar(self, param):
        return "TEXT"

    def convert(self, value, param, ctx):
        if value is None:
            return None
        elif value == "":
            return EXPLICIT_NULL
        else:
            return value


class UrlOrNull(StringOrNull):
    """
    Very similar to StringOrNull, but validates that the string is parsable as an
    http or https URL.
    """

    def get_metavar(self, param):
        return "TEXT"

    def convert(self, value, param, ctx):
        if value is None:
            return None
        elif value == "":
            return EXPLICIT_NULL
        else:
            try:
                url = urlparse(value)
                assert url[0] in ["http", "https"]
            except Exception:
                raise click.UsageError(
                    "{} did not contain a well-formed http or https URL".format(value)
                )
            return value


class CommaDelimitedList(click.ParamType):
    def get_metavar(self, param):
        return "TEXT,TEXT,..."

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        if value is None:
            return None
        # if `--foo` is a comma delimited list and someone passes
        # `--foo ""`, take that as `foo=[]` rather than foo=[""]
        #
        # the alternative is fine, but we have to choose one and this is
        # probably "closer to what the caller meant"
        #
        # it means that if you take
        # `--foo={",".join(original)}`, you will get a value equal to
        # `original` back if `original=[]` (but not if `original=[""]`)
        elif value == "":
            return []
        else:
            return value.split(",")


class _StringPrefixMapper(click.ParamType):
    """
    This is the base class for mapping types which try to split up inputs and parse them
    based on identifying prefixes

    It can be used to define a nice system for dispatch on prefixes
    """

    __prefix_mapping__ = {}
    __prefix_metavars__ = []

    def __init__(self, *args, null=None, **kwargs):
        self.null = null
        super().__init__(*args, **kwargs)

    def get_metavar(self, param):
        return "[" + "|".join(self.__prefix_metavars__) + "]"

    def convert(self, value, param, ctx):
        if value is None:
            return None
        if self.null is not None and value == self.null:
            return EXPLICIT_NULL

        return self.prefix_mapper_parse_input(value)

    def _prefix_mapper_get_parser(self, parsername):
        return getattr(self, parsername)

    def prefix_mapper_default_parser(self, value):
        """override-able default (no-op)"""
        return value

    def prefix_mapper_parse_input(self, value):
        """
        Given an input, try to map it to a parsing func by prefix
        """
        for prefix, parser in self.__prefix_mapping__.items():
            if value.startswith(prefix):
                value = value[len(prefix) :]
                return self._prefix_mapper_get_parser(parser)(value)
        return self.prefix_mapper_default_parser(value)


class JSONStringOrFile(_StringPrefixMapper):
    """
    This type parses a JSON string or falls back to loading JSON data from a
    path, specified by `file:PATH`

    Implements file: -> json
    Default is parse-as-JSON
    """

    __prefix_mapping__ = {"file:": "prefix_mapper_parse_json_file"}
    __prefix_metavars__ = ["JSON", "file:JSON_FILE"]

    def prefix_mapper_parse_json_file(self, value):
        try:
            with open(value, "r") as fp:
                return json.load(fp)
        except json.JSONDecodeError:
            raise click.UsageError("{} did not contain valid JSON".format(value))
        except FileNotFoundError:
            raise click.UsageError("FileNotFound: {} does not exist".format(value))

    def prefix_mapper_default_parser(self, value):
        """
        The mapper also provides a shared JSON string parser which produces nice errors
        """
        # try to handle by parsing as JSON data
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            addendum = (
                ". Did you mean to use 'file:'?" if value.startswith("file") else ""
            )
            # did not match as a URI or parse as JSON, error
            raise click.UsageError(
                "the string '{}' is not valid JSON{}".format(value, addendum)
            )


class IdentityMappingType(JSONStringOrFile):
    """
    This type parses an IdentityMapping input as a string. It can handle
    ExternalIdentityMappings smoothly, but simply accepts JSON input to deal with the
    ExpressionIdentityMapping case

    It inherits JSON handling from JSONStringOrFile

    Inputs are strings be of one of the following forms

      file:X
        X is a file containing JSON data for a full IdentityMapping object
      external:X
        X is a command which should be run for an ExternalIdentityMapping
        it *will* be reparsed by the CLI with shlex, so that it can be passed as a list
        in the ExternalIdentityMapping
      X  (fallback)
        X is a JSON string containing a full IdentityMapping object
    """

    __prefix_mapping__ = {
        "file:": "prefix_mapper_parse_json_file",
        "external:": "prefix_mapper_parse_external_command",
    }
    __prefix_metavars__ = ["external:CMD", "file:JSON_FILE", "JSON"]

    def prefix_mapper_parse_external_command(self, value):
        command = shlex.split(value)
        return {"DATA_TYPE": "external_identity_mapping#1.0.0", "command": command}


class NetworkPortRange(click.ParamType):
    def get_metavar(self, param):
        return "[LOW_PORT,HIGH_PORT|LOW_PORT-HIGH_PORT]"

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        if value is None:
            return None

        # determine the delimiter to use, handling malformed things on the way
        delim = None
        if "," not in value and "-" not in value:
            self.fail("port ranges must be separated with either ',' or '-'")
        elif "," in value and "-" in value:
            self.fail("port ranges cannot contain both ',' and '-'")
        elif "," in value:
            delim = ","
        elif "-" in value:
            delim = "-"
        # split and verify that it's exactly two values
        value = value.split(delim)
        if len(value) != 2:
            self.fail("port ranges must have exactly two values")
        # convert the values to integers
        try:
            value = int(value[0]), int(value[1])
        except ValueError:
            self.fail("port range failed to parse as two integers")
        # make sure the integers are in the 1-65,535 range
        for num in value:
            if 65535 < num:
                self.fail("port numbers cannot exceed 65,535")
            if 1 > num:
                self.fail("port numbers must be positive")
        if value[0] > value[1]:
            self.fail("low port cannot be higher than high port")
        return value
