import json
import warnings

from globus_sdk.exc import GlobusAPIError
from globus_sdk.response import GlobusHTTPResponse


class GCSResponse(GlobusHTTPResponse):
    """
    Base class for :class:`GCSClient <globus_sdk.GCSClient>` responses
    """

    def __init__(self, data, client=None):
        # check result version as we make several assumptions about its format
        super(GCSResponse, self).__init__(data, client=client)
        try:
            data_type = self._data.json()["DATA_TYPE"]
            if data_type != "result#1.0.0":
                warnings.warn(
                    "Unexpected GCS API Result version {}. It is recommended "
                    "that you update your client software if possible.".format(
                        data_type
                    )
                )
        except KeyError:
            warnings.warn(
                "Unknown GCS API Result version. It is recommended that you "
                "update your client software if possible."
            )

    def __str__(self):
        # Make printing responses more convenient. Relies on the
        # fact that GCS API responses are always JSON.
        return json.dumps(self.data, indent=2)


class SingletonGCSResponse(GCSResponse):
    """
    Response class for resources that should return exactly one item of
    response data
    """

    @property
    def data(self):
        """
        Response data as a Python data structure. Usually a dict or
        list.
        """
        try:
            data_len = len(self._data.json()["data"])
            if data_len != 1:
                warnings.warn(
                    "GCSClient expected one item of response data, but found "
                    "{}. Top level field access will be used".format(data_len)
                )
                return self._data.json()
            else:
                return self._data.json()["data"][0]
        except KeyError:
            warnings.warn(
                "GCSClient expected one item of response data, but found "
                "no data field. Top level field access will be used."
            )
            return self._data.json()


class IterableGCSResponse(GCSResponse):
    """
    Response class for non-paged list oriented resources. Allows top level
    fields to be accessed normally via standard item access, and also
    provides a convenient way to iterate over the sub-item list in the
    ``DATA`` key:

    >>> print("Path:", r["path"])
    >>> # Equivalent to: for item in r["DATA"]
    >>> for item in r:
    >>>     print(item["name"], item["type"])
    """

    def __iter__(self):
        return iter(self["data"])


class GCSAPIError(GlobusAPIError):
    """
    Error class for the GCS API client
    """

    def __init__(self, *args, **kwargs):
        self.json_data = None
        self.detail_data_type = None
        super().__init__(*args, **kwargs)

    def _load_from_json(self, data):
        self.json_data = data
        self.code = data["code"]
        self.message = data.get("message")

        # detail can be a full document, so fetch, then look for a DATA_TYPE
        # and expose it as a top-level attribute for easy access
        self.detail = data.get("detail")
        if self.detail and "DATA_TYPE" in self.detail:
            self.detail_data_type = self.detail["DATA_TYPE"]
