from flask import current_app

from globus.manager.api.errors import AuthorizationError, NotResourceOwner
from .authentication_timeout_filter import AuthenticationTimeoutFilter
from .identity_set import IdentitySet
from .identity_in_identity_set_filter import IdentityInIdentitySetFilter


class IdentityMatchesResourceOwnerFilter:
    def __init__(self, resource_owner):
        self.resource_owner = resource_owner

    def __call__(self, identity_set: IdentitySet, auth_state):
        """
        Filter identity_set, removing all items which do not
        have the identity_id associated with this filter.

        :param identity_set:
            Identity information to filter
        :param auth_state:
            Authentication state for this request

        :returns: Returns the subset of identity_set_data which matches
            the identity_id.
        """
        invalid_identities = [i for i in identity_set if i["id"] != self.resource_owner]
        for i in invalid_identities:
            identity_set.remove(i)

        if invalid_identities:
            detail = NotResourceOwner(self.resource_owner)
            identity_set.error_details.insert(0, detail)

        return identity_set


class CredentialOwnerAuthorizer:
    """
    This class provides methods to process the storage gateway authorization
    policies
    """

    def __init__(self, credential):
        self.filters = []

        self.filters.append(IdentityMatchesResourceOwnerFilter(credential.identity_id))

        sg = credential.storage_gateway
        self.filters.append(
            AuthenticationTimeoutFilter(
                require_session=sg.high_assurance,
                authentication_timeout=(sg.authentication_timeout_mins),
            )
        )

    def authorize_identity(self, identity_id, auth_state):
        """
        Return an entry from identity_set_data which matches the owner
        of this credential
        """
        idset = IdentitySet(auth_state.identity_set_data)
        f = IdentityInIdentitySetFilter(identity_id)
        f(idset, auth_state)

        for f in self.filters:
            f(idset, auth_state)

        if idset:
            assert len(idset) == 1
            return idset[0]

        detail = None
        if idset.error_details:
            detail = idset.error_details[0]
        raise AuthorizationError(detail=detail)

    def authorize_identities(self, auth_state):
        """
        Return a list of entries from identity_set_data which match the
        storage gateway policies.
        """
        identities = IdentitySet()
        for i in auth_state.identity_set_data:
            try:
                identity = self.authorize_identity(i["id"], auth_state)
                identities.append(identity)
            except Exception as e:
                current_app.logger.debug(
                    "Denied authorization for %s: %s", i["id"], str(e)
                )
                identities.error_details.append(e.detail)

        return identities
