from flask import current_app

from globus.manager.api.models import Endpoint

from globus.manager.api.errors import AuthorizationError, NotPublic
from .identity_set import IdentitySet
from .identity_in_identity_set_filter import IdentityInIdentitySetFilter


class PublicEndpointFilter:
    def __init__(self):
        endpoint = Endpoint.query.one_or_none()
        self.public = endpoint and endpoint.public

    def __call__(self, identity_set, auth_state):
        """
        Filter the identity_set, removing all items if the public
        property of this filter is False.

        :param identity_set:
            Identity set information, filtered in place
        :param auth_state:
            Authentication state for this request

        :returns: Returns either the original identity_set or empties it
            and sets an error detail.
        """
        invalid_identities = [i for i in identity_set if not self.public]

        for i in invalid_identities:
            identity_set.remove(i)

        if invalid_identities or not identity_set:
            detail = NotPublic()
            identity_set.error_details.insert(0, detail)

        return identity_set


class PublicAuthorizer:
    """
    This class provides methods to allow access to a caller if the
    endpoint is public.
    """

    def __init__(self):
        self.filters = [PublicEndpointFilter()]

    def authorize_identity(self, identity_id, auth_state):
        """
        Return an entry from identity_set_data which is allowed to
        access this resource.
        """
        idset = IdentitySet(identity_set=auth_state.identity_set_data)

        # Filter out identities which don't match the requested id first
        f = IdentityInIdentitySetFilter(identity_id)
        f(idset, auth_state)

        for f in self.filters:
            f(idset, auth_state)

        if idset:
            assert len(idset) == 1
            return idset[0]

        detail = None
        if idset.error_details:
            detail = idset.error_details[0]

        raise AuthorizationError(detail=detail)

    def authorize_identities(self, auth_state):
        """
        Return a list of entries from the auth state's identity_set_data which
        match the storage gateway policies.
        """
        idset = IdentitySet()
        for i in auth_state.identity_set_data:
            try:
                self.authorize_identity(i["id"], auth_state)
                idset.append(i)
            except Exception as e:
                current_app.logger.debug(e)

        return idset
