import functools
from typing import Callable, Optional

import base62
from flask import current_app, g, request
from werkzeug.exceptions import HTTPException

import marshmallow as ma
from globus.manager.api.errors import (
    BadRequestError,
    ValidationError,
    UnsupportedMediaError,
)
from globus.manager.api.utils.server_timing import ServerTiming


def expects_json(
    fn: Optional[Callable] = None, schema: Optional[ma.Schema] = None,
):
    """
    Decorate a route to require a JSON body and optionally validate against a
    marshmallow schema.

    :param fn:
        View method to execute if the expectation is met. When called, this
        decorator will populate `g.validated_data`. If there is no schema
        associated with this decorator, the value of `g.validated_data` is
        set to `None`, otherwise it is the output of deserializing the input
        object with the marshmallow schema.
    :param schema:
        A schema instance to use to validate the input.

    :raises UnsupportedMediaError:
        The input content-type is not application/json
    :raises BadRequestError:
        The input could not be loaded from the HTTP request.
    :raises ValidationError:
        The input does not validate against the schema.
    """
    if fn is None:
        return functools.partial(expects_json, schema=schema)

    @functools.wraps(fn)
    def decorated_function(*args, **kwargs):
        if not request.is_json:
            raise UnsupportedMediaError()

        try:
            json = request.get_json()
        except HTTPException as ex:
            raise BadRequestError(ex.description)

        g.validated_data = None

        if schema:
            try:
                g.validated_data = schema.load(json)
                current_app.logger.debug(
                    "Validated Data ({}): {}".format(
                        schema.__class__.__name__ if schema else "any",
                        g.validated_data,
                    )
                )
            except ma.ValidationError as e:
                current_app.logger.error(
                    "deserialization error: {} {} using schema {}".format(
                        e, json, schema
                    )
                )
                raise ValidationError(exc=e)

        return fn(*args, **kwargs)

    return decorated_function


# Server metrics we want to provide to API consumers
db_read = ("db-read", "Database Reads")
db_write = ("db-write", "Database Writes")
transfer_api = ("t-api", "Transfer API")
auth_api = ("auth-api", "Auth API")
letsencrypt = ("letsencrypt", "LetsEncrypt")


def server_timing(fn):
    @functools.wraps(fn)
    def decorated_function(*args, **kwargs):
        g.server_timing = ServerTiming()

        return fn(*args, **kwargs)

    return decorated_function


class Base62Marker:
    """
    Generic safe string marker implementation. This is used as the
    default marker class for the paginated decorator.
    """

    def __init__(self, marker=None, marker_data=None):
        """
        Initializer for the Base62Marker class. One of the two named parameters
        can be used to initialize the marker.

        :param marker:
            The base62 encoded representation of this marker. This is typically
            used when parsing the marker query parameter to a route.
        :param marker_data:
            A string version of the data which will be stored in the marker.
        """
        if marker:
            self.decode(marker)
            if marker_data:
                raise ValueError("Invalid marker")
        else:
            self.marker = marker_data

    @property
    def marker(self):
        """
        The string version of the data stored in the marker.
        """
        return self._marker

    @marker.setter
    def marker(self, value):
        self._marker = value

    def decode(self, marker):
        """
        Decode the base62 marker string and store its value in the
        `marker` property
        """
        self.marker = base62.decodebytes(marker).decode("utf-8")

    def encode(self):
        """
        Encode and return the `marker` property as a base62 string.
        """
        return base62.encodebytes(self.marker.encode("utf-8"))

    def __str__(self):
        return self.marker


def paginated(fn=None, marker_class=None):
    """
    Decorate a route to parse and validate pagination parameters `marker` and
    `page_size`. The results of parsing are stored in `g.marker` and
    `g.page_size`.

    :param fn:
        The route handler to call after parsing the optional pagination
        parameters.
    :param marker_class:
        An optional class that implements a constructor that can be passed a
        `marker` named parameter with encode() and decode() methods to parse
        and serialize the marker. If not present or None, the
        :py:class:`globus.manager.api.blueprints.decorators.Base62Marker` class
        is used.
    """
    if fn is None:
        return functools.partial(paginated, marker_class=marker_class)

    if marker_class is None:
        marker_class = Base62Marker

    @functools.wraps(fn)
    def decorated_function(*args, **kwargs):
        page_size_string = request.args.get("page_size", "16")
        marker_string = request.args.get("marker")
        page_size = None
        marker = None

        try:
            page_size = int(page_size_string)
            if page_size < 1 or page_size > 100:
                raise ValueError(
                    "The page_size parameter must be a positive value between "
                    "1 and 100"
                )
        except ValueError:
            raise BadRequestError("Invalid page_size parameter")

        if marker_string:
            try:
                marker = marker_class(marker=marker_string)
            except (ValueError, AttributeError, UnicodeDecodeError):
                raise BadRequestError("Invalid marker")

        g.page_size = page_size
        g.marker = marker
        return fn(*args, **kwargs)

    return decorated_function
