from flask import Blueprint, current_app, g, jsonify, request

import globus.manager.constants as c
from globus.manager.api.schemas.api import (
    collection_private_schema,
    result_schema,
    storage_gateway_private_schema,
    user_credential_private_schema,
    path_restrictions_schema,
)
from globus.manager.api.schemas.gridftp.gridftp_login import gridftp_login_schema
from globus.manager.api.authentication.authstate import AuthenticationState
from globus.manager.api.authentication.decorators import authenticated_with_secret
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.blueprints.decorators import expects_json
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    put_object,
    update_object,
)
import globus.manager.api.models as m
import globus.manager.api.errors as errors
from globus.manager.api.models.result import Result


blueprint = Blueprint("private_gridftp_blueprint", __name__, url_prefix="/api")


@blueprint.route("/private/gridftp_login/<uuid:collection_id>", methods=["POST"])
@authenticated_with_secret
@expects_json(schema=gridftp_login_schema)
def private_gridftp_login_post(collection_id):
    """
    POST /private/gridftp_login

    Processes a user login for a collection, returning information about the
    collection, storage_gateway, and user_credential as requested.
    """
    includes = request.args.get("include", "").split(",")

    collection = m.Collection.query.filter(
        ~m.Collection.deleted, m.Collection.id == str(collection_id)
    ).one_or_none()

    if not collection:
        raise errors.NotFoundError()

    if collection:
        check_subscription(
            collection.storage_gateway.connector,
            high_assurance=collection.storage_gateway.high_assurance,
            collection_type=c.CollectionType(collection.collection_type),
        )

        if collection.is_guest_collection:
            if (
                collection.mapped_collection.deleted
                or not collection.mapped_collection.allow_guest_collections
            ):
                raise errors.NotFoundError("No such collection")

    storage_gateway = collection.storage_gateway
    user_credential = None

    if storage_gateway.deleted:
        raise errors.NotFoundError()

    current_app.logger.info(
        "login: %s identities: %s", collection, g.validated_data["identity_ids"]
    )

    identity_set_data = []

    if c.CollectionType(collection.collection_type) == c.CollectionType.guest:
        # For a guest collection, we will process the API based on the
        # identity_id information in the collection's user_credential if
        # present.
        if collection.user_credential:
            identity_set_data = current_app.auth_client.get_identities(
                ids=[collection.user_credential.identity_id]
            )["identities"]
        else:
            identity_set_data = current_app.auth_client.get_identities(
                ids=[collection.identity_id]
            )["identities"]
    elif c.CollectionType(collection.collection_type) == c.CollectionType.mapped:
        # For a mapped collection, we process the API based on the
        # identity_id information passed to this API.
        if g.validated_data["identity_ids"]:
            identity_set_data = current_app.auth_client.get_identities(
                ids=[str(id) for id in g.validated_data["identity_ids"]]
            )["identities"]

    # Synthesize some authentication state based on the identities
    # we've received from the GridFTP server. Those were verified
    # by Transfer at the time of the job submission, or by HTTPS
    # by introspecting the auth tokens.
    auth_state = AuthenticationState(identity_set_data=identity_set_data)

    # Because the authentication timeouts were validated at the
    # time the job was submitted from transfer, we don't verify
    # them now.
    authorizer = StorageGatewayAuthorizer(
        storage_gateway, enforce_authentication_timeout=False
    )
    authorized_identities = authorizer.authorize_identities(auth_state)

    if not authorized_identities:
        detail = None
        if authorized_identities.error_details:
            detail = authorized_identities.error_details[0]
        raise errors.AuthorizationError(detail=detail)

    # Map the authorized identity set to allowed local usernames if any
    mappings = collection.storage_gateway.map_identities(authorized_identities)

    if not mappings:
        raise errors.InvalidUserError()

    for map_result in mappings:
        if collection.user_credential:
            # For a guest collection, the collection will have a
            # specific user_credential associated with it. If it matches
            # our identity we can use it
            if collection.user_credential.identity_id == map_result.identity_id:
                if collection.user_credential.username == map_result.username:
                    user_credential = collection.user_credential
                    break
        else:
            # For a mapped collection, we will see if the user has
            # registered a user_credential on the storage_gateway. If the
            # Storage Gateway maps identities, then make sure the
            # credential maps to that identity, otherwise, choose the
            # credential that is registered
            user_credential = m.UserCredential.query.filter(
                ~m.UserCredential.deleted,
                m.UserCredential.storage_gateway_id == storage_gateway.id,
                m.UserCredential.identity_id == map_result.identity_id,
                m.UserCredential.username == map_result.username,
            ).one_or_none()
        if user_credential:
            break

    if collection.user_credential and not user_credential:
        # This collection has a user credential, but none of the mappings are
        # for that credential. This means that an identity has changed since
        # the credential was created and it is no longer usable
        raise errors.InvalidCredentialError(
            "This collection was created with a credential which is "
            "no longer accessible by the account which created it",
            user_credential_id=collection.user_credential.id,
        )

    if user_credential and not user_credential.invalid:
        if user_credential.needs_refresh():
            try:
                current_app.logger.error("Refreshing credential")
                user_credential.refresh()
            except errors.APIBaseError:
                raise
            except Exception as e:
                raise errors.InternalServerError(
                    "Unable to refresh credential: {}".format(e)
                )
            try:
                update_object(user_credential, lambda o: o)
            except Exception as e:
                current_app.logger.warning(
                    "Error refreshing %s: %s", user_credential, str(e)
                )
                m.db.session.rollback()

                raise errors.InvalidCredentialError("Unable to refresh user credential")

        # Raises an error if the credential can not be used for this collection
        try:
            collection.check_user_credential(user_credential)
        except errors.APIBaseError as e:
            if user_credential.invalid:

                def mark_invalid(cred):
                    cred.invalid = True

                update_object(user_credential, mark_invalid)
                g.updated_objects = user_credential.id
            raise e

    data = []

    if not user_credential or user_credential.invalid:
        if user_credential:
            deleted_user_credential = user_credential
            authorized_identity = next(
                ai
                for ai in authorized_identities
                if ai["id"] == user_credential.identity_id
            )
            username = user_credential.username
        elif len(mappings) == 1:
            authorized_identity = next(
                ai
                for ai in authorized_identities
                if ai["id"] == mappings[0].identity_id
            )
            # If there's an existing credential, we'll need to update
            # instead of create in appsync
            deleted_user_credential = m.UserCredential.query.filter(
                m.UserCredential.deleted,
                m.UserCredential.storage_gateway_id == storage_gateway.id,
                m.UserCredential.identity_id == authorized_identity["id"],
                m.UserCredential.username == mappings[0].username,
            ).one_or_none()
        elif collection.storage_gateway.require_credential:
            raise errors.CredentialNotFoundError(
                accounts=[mapping for mapping in mappings]
            )

        # If there is no user credential in the db, and there is only
        # one mapping for the user, create one on the fly. This may be
        # marked as an invalid credential for connectors that require
        # some intervention to populate the credential
        uc_doc = {
            "DATA_TYPE": "user_credential#1.0.0",
            "identity_id": authorized_identity["id"],
            "username": mappings[0].username,
            "storage_gateway_id": storage_gateway.id,
        }
        if deleted_user_credential:
            deleted_user_credential.deleted = False
            put_object(deleted_user_credential, uc_doc)
            user_credential = deleted_user_credential
        else:
            cls = m.UserCredential.polymorphic_class(storage_gateway.connector_id)
            current_app.logger.error("Creating new credential object from %s", uc_doc)
            user_credential = cls(doc=uc_doc, storage_gateway=storage_gateway,)
            current_app.logger.error(
                "Creating new credential object %s", user_credential
            )
            create_object(user_credential)

    if not user_credential:
        raise errors.CredentialNotFoundError(accounts=[mapping for mapping in mappings])
    elif user_credential.invalid:
        if c.CollectionType(collection.collection_type) == c.CollectionType.guest:
            raise errors.InvalidCredentialError(
                "This collection is unusable, as it is configured with an invalid "
                "credential. Please contact the collection owner to update "
                "its configuration."
            )
        elif user_credential.provisioned:
            raise errors.InvalidCredentialError(user_credential_id=user_credential.id)
        else:
            raise errors.InvalidCredentialError(
                "Your credential requires some initial setup.",
                user_credential_id=user_credential.id,
            )

    username = user_credential.username

    if "user_credential" in includes:
        data.append(user_credential_private_schema.dump(user_credential))

    if "collection" in includes:
        collection_data = collection_private_schema.dump(collection)
        if "sharing_restrict_paths" in collection_data:
            collection_data["sharing_restrict_paths"] = path_restrictions_schema.dump(
                collection.absolute_sharing_restrict_paths(username)
            )
        collection_data["root_path"] = collection.absolute_root_path(username)
        data.append(collection_data)

    if "storage_gateway" in includes:
        storage_gateway_data = storage_gateway_private_schema.dump(storage_gateway)
        if "restrict_paths" in storage_gateway_data:
            storage_gateway_data["restrict_paths"] = path_restrictions_schema.dump(
                storage_gateway.absolute_restrict_paths(username=username)
            )

        data.append(storage_gateway_data)

    res = result_schema.dump(Result(message="Login ok", data=data))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
