from flask import (
    Blueprint,
    current_app,
    g,
    jsonify,
    request,
)

import sqlalchemy

from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
)
from globus.manager.api.blueprints.utils import (
    create_object,
    delete_object,
)
import globus.manager.api.errors as errors
from globus.manager.api.models import (
    Collection,
    ConfigurationObject,
    Role,
    Result,
)
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.role import role_schema
from globus.manager.api.authorization import requires_one_role, requires_subscription
from globus.manager.constants import RoleType


blueprint = Blueprint("roles", __name__, url_prefix="/api/")


def get_collection_from_collection_id_query():
    """
    Resolves the collection_id query parameter to a Collection
    model for authorization
    """
    collection_id = request.args.get("collection_id")
    if collection_id:
        return Collection.query.filter(
            ~Collection.deleted, Collection.id == collection_id
        ).one_or_none()
    else:
        return None


def get_collection_from_role_object():
    """
    Resolves the collection_id from the parsed input object
    to a Collection model for authorization
    """
    collection_id = g.validated_data.get("collection")
    if collection_id:
        try:
            g.collection = Collection.query.filter(
                ~Collection.deleted, Collection.id == str(collection_id)
            ).one()
            return g.collection
        except sqlalchemy.orm.exc.NoResultFound as e:
            current_app.logger.info(e)
            raise errors.ForbiddenError()


def get_collection_from_role_id_path():
    """
    Resolves the collection_id to the Collection if the
    path parameter role_id resolves to a Role assignment for
    a collection.
    """
    role_id = request.view_args.get("role_id")
    if role_id:
        role_id = str(role_id)
        role = Role.query.filter(~Role.deleted, Role.id == role_id).one_or_none()
        if role and role.resource:
            collection = Collection.query.filter(
                ~Collection.deleted, Collection.id == role.resource
            ).one_or_none()
            return collection
    return None


@blueprint.route("/roles", methods=["GET"])
@authenticated
@paginated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_from_collection_id_query,
)
def roles_get():
    """
    Get the Globus Connect Server Endpoint's or a Collection's
    list of `Role` documents.

    If the `collection_id` query
    parameter is passed to this operation, then the roles related
    to that Collection are returned.

    The `include` parameter determines whether this
    operation returns all roles relevant to the resource or only
    those that the caller has.

    To obtain information about *all* roles, the caller must pass
    the "all_roles" value as the value of the "include" parameter.
    This requires the "administrator" role for the
    Endpoint, or Collection the role is associated with.
    """
    includes = request.args.get("include", "").split(",")
    marker = g.marker
    page_size = g.page_size

    collection_id = request.args.get("collection_id")
    include_all_roles = "all_roles" in includes

    if collection_id is not None:
        role_query = Role.query.filter(
            ~Role.deleted, Role.is_collection_role, Role.resource == collection_id
        )
    else:
        role_query = Role.query.filter(~Role.deleted, Role.is_endpoint_role)

    # TODO: Handle group role assignments
    if not include_all_roles:
        role_query = role_query.filter(
            Role.identity_id.in_([i["id"] for i in g.auth_state.identity_set_data])
        )

    if marker:
        role_query = role_query.filter(ConfigurationObject.id >= marker.marker)

    if not current_app.subscription:
        role_query = role_query.filter(Role.role == RoleType.endpoint_owner)

    role_query = role_query.order_by(ConfigurationObject.id).limit(page_size + 1)

    roles = []

    for role in role_query:
        roles.append(role_schema.dump(role))

    if len(roles) > page_size:
        next_page_token = Base62Marker(marker_data=roles.pop()["id"])
    else:
        next_page_token = None

    res = result_schema.dump(Result(data=roles, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/roles", methods=["POST"])
@authenticated
@expects_json(schema=role_schema)
@requires_subscription
@requires_one_role(
    role_names=["endpoint:administrator", "collection:administrator:{collection_id}"],
    require_session=True,
    authentication_timeout=None,
    collection_selector=get_collection_from_role_object,
)
def roles_post():
    """
    Assign a role to an identity or group for the Endpoint or a
    specified collection based on the collection_id query parameter.

    On success returns a copy of the created role with the system
    generated id added.
    """
    # Ignore input id
    g.validated_data.pop("id", None)

    input_role = g.validated_data
    role_collection_uuid = input_role.get("collection")
    if role_collection_uuid:
        role_collection = str(role_collection_uuid)
        if g.collection.is_mapped_collection and input_role["role"] == "access_manager":
            raise errors.BadRequestError("Invalid role for mapped collection")
    else:
        role_collection = None
        if input_role["role"] == "access_manager":
            raise errors.BadRequestError("Invalid role for endpoint")

    if input_role["role"] == "owner":
        raise errors.BadRequestError("Owner role may not be created explicitly")

    # Validate the principal if it is a identity. Auth does not have a proper
    # groups API for validating groups; you can only validate groups you belong
    # to.
    principal = input_role.get("principal")
    identity_urn_prefix = "urn:globus:auth:identity:"
    if principal.startswith(identity_urn_prefix):
        auth_identity = current_app.auth_client.get_identities(
            ids=[principal[len(identity_urn_prefix) :]]
        ).data.get("identities", [])
        if len(auth_identity) != 1 or auth_identity[0]["status"] != "used":
            raise errors.BadRequestError("Role principal is not a valid Globus account")

    with current_app.appsync_client.pause_notifications():
        if Role.query.filter(
            ~Role.deleted,
            Role.principal == input_role["principal"],
            Role.role_name == input_role["role"],
            Role.resource.is_(role_collection),
        ).one_or_none():
            raise errors.ResourceExistsError()

        try:
            role = Role(doc=input_role)
        except errors.APIBaseError:
            raise
        except Exception as e:
            current_app.logger.info("Unable to create role: %s", str(e))
            raise errors.InternalServerError("Unable to create role")

        create_object(role)

        res = result_schema.dump(
            Result(
                message="Created new role {}".format(role.id),
                data=[role_schema.dump(role)],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.created_objects.append(role.id)

    return response


@blueprint.route("/roles/<uuid:role_id>", methods=["DELETE"])
@authenticated
@requires_subscription
@requires_one_role(
    role_names=["endpoint:administrator", "collection:administrator:{collection_id}"],
    collection_selector=get_collection_from_role_id_path,
)
def roles_delete(role_id):
    """
    Delete one of the Globus Connect Server's `Role`
    documents by its ID
    """

    current_app.logger.debug("DELETE role {}".format(role_id))
    role_id = str(role_id)

    with current_app.appsync_client.pause_notifications():
        role = Role.query.filter(~Role.deleted, Role.id == role_id).one_or_none()

        if not role:
            raise errors.NotFoundError(
                error_description="Role {} not found".format(role_id)
            )

        if role.primary_role:
            raise errors.ConflictError("Unable to delete primary role")

        delete_object(role)

        res = result_schema.dump(Result(message="Deleted role {}".format(role_id)))

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.deleted_objects.append(role.id)

    return response


@blueprint.route("/roles/<uuid:role_id>", methods=["GET"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_from_role_id_path,
)
def roles_get_by_id(role_id):
    """
    Get one of the Globus Connect Server's `Role`
    documents by its ID
    """

    role_id = str(role_id)
    role = Role.query.filter(~Role.deleted, Role.id == role_id).one_or_none()
    if not role:
        raise errors.NotFoundError()

    res = result_schema.dump(Result(data=[role_schema.dump(role)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
