from flask import current_app
from typing import Optional

import globus_sdk

from globus.manager.api.models import (
    Connector,
    PosixStorageGateway,
    db,
)
from globus.manager.appsync.query import GraphQLError
import globus.manager.api.errors as errors
from globus.manager.constants import CollectionType


def check_subscription(
    connector: Connector,
    high_assurance: bool = False,
    collection_type: Optional[CollectionType] = None,
):
    """
    Perform a subscription check to create a resource using the given
    storage gateway and/or collection type. If either are missing, they
    do not contribute to the required subscription level and addons.

    By default, endpoint operations requires only the "basic" level of
    subscription, but more is required for the following cases:

    -   Guest collections require at least a starter level subscription
    -   Premium connectors require at least a standard level subscription
    -   High assurance storage gateways and collections require at least a
        high_assurance level subscription

    Additionally, when a user attempts to create a resource using a premium
    connector, they will need an add-on for the connector in addition to the
    standard subscription.

    :param connector:
        The connector being checked against the endpoint's subscription.
    :param high_assurance:
        An optional flag indicating whether this check is being done to
        create a high assurance storage gateway. This value is only checked
        if the `storage_gateway` parameter is a class type.
    :param collection_type:
        An optional collection type enum value indicating whether this is
        evaluating the ability to access or create a mapped or a guest
        collection.

    :raises errors.SubscriptionRequiredError:
        If the storage_gateway or collection type indicate a non-basic
        level of subscription is needed, this function raises the
        SubscriptionRequiredError which will indicate what features are
        required in the subscription.
    """
    subscription_level = "basic"
    addons = []

    if collection_type == CollectionType.guest:
        subscription_level = "starter"

    if connector.id != PosixStorageGateway.CONNECTOR_ID:
        addons.append(connector.display_name)
        subscription_level = "standard"

    if high_assurance:
        subscription_level = "standard"
        addons.append("high_assurance")

    # XXX: Add support for subscription service to check properties of the
    # subscription. Right now, assume a subscription allows everything
    if not current_app.subscription and subscription_level != "basic":
        raise errors.SubscriptionRequiredError(
            subscription_level=subscription_level, add_ons=addons
        )


def update_object(obj, update_callable, *args, **kwargs):
    """
    Update the given object using update_callable (called with `obj`, `*args`,
    `**kwargs`), then update it in the database and appsync.
    """
    committed_to_appsync = False
    retries_remaining = 5

    while not committed_to_appsync and retries_remaining > 0:
        try:
            update_callable(obj, *args, **kwargs)
            db.session.add(obj)
        except (ValueError, KeyError, AttributeError) as e:
            current_app.logger.error("UPDATE: %s %s %s", str(obj), type(e), str(e))
            obj.rollback()
            db.session.rollback()
            raise errors.ValidationError(str(e))
        except errors.APIBaseError as e:
            current_app.logger.error("UPDATE: %s %s", str(obj), str(e))
            obj.rollback()
            db.session.rollback()
            raise
        except Exception as e:
            current_app.logger.error("UPDATE: %s %s %s", str(obj), type(e), str(e))
            db.session.rollback()
            obj.rollback()
            raise errors.InternalServerError(str(e))

        try:
            current_app.appsync_client.update_object(obj)
            committed_to_appsync = True
        except GraphQLError as e:
            current_app.logger.warning("UPDATE: %s: %s", str(obj), e.message)  # noqa
            if e.errors[0].get("errorType") == "CONFLICT":
                if e.data is None:
                    # deleted
                    db.session.rollback()
                    obj.rollback()
                    raise errors.NotFoundError()
                else:
                    # object changed before our patch was applied
                    db.session.rollback()
                    obj.rollback(e.data)
                    retries_remaining -= 1
            else:
                # other GraphQL or transmission error
                retries_remaining -= 1
        except Exception as e:
            # Unexpected error
            db.session.rollback()
            obj.rollback()
            current_app.logger.error("UPDATE %s: %s:", str(obj), str(e))
            raise errors.BadGatewayError()

    if not committed_to_appsync:
        db.session.rollback()
        obj.rollback()
        raise errors.ConflictError()

    db.session.commit()


def patch_object(obj, patch_document):
    """
    Patch the given object, update it in the database and appsync.
    """
    return update_object(obj, lambda o: o.patch(**patch_document))


def put_object(obj, put_document):
    """
    Put the given object, update it in the database and appsync.
    """
    return update_object(obj, lambda o: o.put(**put_document))


def create_object(obj):
    """
    Publish the given object to appsync and commit it to the database
    """
    committed_to_appsync = False
    retries_remaining = 5

    while not committed_to_appsync and retries_remaining > 0:
        try:
            current_app.appsync_client.create_object(obj)
            committed_to_appsync = True
        except GraphQLError as e:
            current_app.logger.warning("POST: %s: %s", obj.id, e.message)  # noqa
            if e.errors[0].get("errorType") == "CONFLICT":
                obj.rollback()
                db.session.rollback()
                raise errors.ResourceExistsError()
            else:
                retries_remaining -= 1
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.warning("POST: %s: %s", obj.id, e.message)  # noqa
            retries_remaining -= 1
        except Exception as e:
            obj.rollback()
            db.session.rollback()
            current_app.logger.error("POST: %s: %s", str(obj), str(e))
            raise errors.BadGatewayError()

    if not committed_to_appsync:
        current_app.logger.error(
            "POST: %s: failed to create object in appsync", str(obj)
        )
        obj.delete()
        db.session.delete(obj)
        db.session.commit()
        raise errors.BadGatewayError()

    db.session.add(obj)
    db.session.commit()


def delete_object(obj, *args, update=False, **kwargs):
    """
    Delete the given object and remove it from the database and appsync.
    """
    tag = obj.tag

    committed_to_appsync = False
    retries_remaining = 5

    while not committed_to_appsync and retries_remaining > 0:
        try:
            obj.delete(*args, **kwargs)
        except errors.APIBaseError as e:
            current_app.logger.error("DELETE %s %s: %s", tag.name, obj.id, str(e))
            db.session.rollback()
            raise
        except Exception as e:
            current_app.logger.error("DELETE %s %s: %s", tag.name, obj.id, str(e))
            raise errors.InternalServerError()

        try:
            if update:
                current_app.appsync_client.update_object(obj)
            else:
                current_app.appsync_client.delete_object(obj)
            committed_to_appsync = True
        except GraphQLError as e:
            current_app.logger.warning("DELETE %s: %s", obj.id, e.message)  # noqa
            if e.errors[0].get("errorType") == "CONFLICT":
                if e.data is None:
                    committed_to_appsync = True
                else:
                    obj.rollback(e.data)
                    if obj.deleted:
                        committed_to_appsync = True
                    else:
                        retries_remaining -= 1
            elif e.errors[0].get("errorType") == "NOT_FOUND":
                committed_to_appsync = True
            else:
                retries_remaining -= 1
        except Exception as e:
            obj.rollback()
            db.session.rollback()
            current_app.logger.error("DELETE {} {}: {}".format(tag, obj.id, str(e)))
            raise errors.BadGatewayError()

    if not committed_to_appsync:
        obj.rollback()
        db.session.rollback()
        raise errors.BadGatewayError()

    if update:
        db.session.add(obj)
    else:
        db.session.delete(obj)

    db.session.commit()
