import datetime

from flask import current_app
from sqlalchemy import orm

from .configuration_object import ConfigurationObject
from globus.manager.api.schemas.appsync.acme_account import acme_account_schema
from globus.manager.utils.acme_client import AcmeClient
from globus.manager.constants import ConfigurationObjectTag


class AcmeAccount(ConfigurationObject):
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.ACME_ACCOUNT.name}
    epoch = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)
    delta_zero = datetime.timedelta(seconds=0)
    _DATA_TYPE = "acme_account#1.0.0"

    configuration_schema = acme_account_schema

    def __init__(self, appsync_data=None, doc=None):
        self._lease_node_id = None
        self._lease_expiration = None
        self._acme_key = None
        self._acme_account = None
        self._acme_client = None

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.ACME_ACCOUNT
            args["id"] = ConfigurationObjectTag.ACME_ACCOUNT.name
            super().__init__(**args)
            self.init_on_load()
        elif doc is not None and (doc.get("DATA_TYPE") == self._DATA_TYPE):
            super().__init__(
                tag=ConfigurationObjectTag.ACME_ACCOUNT,
                id=ConfigurationObjectTag.ACME_ACCOUNT.name,
            )
            for k, v in doc.items():
                if k != "DATA_TYPE":
                    setattr(self, k, v)
        elif doc is None:
            super().__init__(
                tag=ConfigurationObjectTag.ACME_ACCOUNT,
                id=ConfigurationObjectTag.ACME_ACCOUNT.name,
            )
            self.init_on_load()
        else:
            raise ValueError("Unknown or missing DATA_TYPE")

    def acquire_lease(self, duration=None):
        self.lease_node_id = current_app.config["NODE_ID"]
        if duration is None:
            duration = datetime.timedelta(hours=12)

        utcnow = datetime.datetime.now(datetime.timezone.utc)

        self.lease_expiration = utcnow + duration

    def lease_remaining(self):
        utcnow = datetime.datetime.now(datetime.timezone.utc)
        try:
            lease_node_id = self.lease_node_id
        except AttributeError:
            lease_node_id = None

        if lease_node_id == current_app.config["NODE_ID"]:
            delta = self.lease_expiration - utcnow

            if delta > self.delta_zero:
                return delta
        return self.delta_zero

    def release_lease(self):
        if self.lease_node_id == current_app.config["NODE_ID"]:
            self.lease_expiration = self.epoch
            self.lease_node_id = None
        else:
            raise ValueError("This node does not have a current lease")

    def get_cert_chain_and_key(self, domain_name):
        current_lease_node_id = getattr(self, "lease_node_id", None)
        utcnow = datetime.datetime.now(datetime.timezone.utc)
        if (
            current_lease_node_id != current_app.config["NODE_ID"]
            or (self.lease_expiration or self.epoch) < utcnow
        ):
            raise ValueError("This node does not have a current lease")

        return self.acme_client.get_cert_chain_and_key(domain_name)

    @orm.reconstructor
    def init_on_load(self):
        if self.configuration_hash:
            acme_config = acme_account_schema.load(
                self.configuration_hash, partial=True
            )
        else:
            acme_config = {}

        self._acme_key = acme_config.get("acme_key")
        self._acme_account = acme_config.get("acme_account")
        self._acme_client = None

        if "lease_node_id" not in self.configuration_hash:
            self._lease_node_id = None
        if "lease_expiration" not in self.configuration_hash:
            self._lease_expiration = None

    @property
    def acme_key(self):
        if self._acme_key is None:
            self._acme_key = self.acme_client.acme_key

        return self._acme_key

    @property
    def acme_account(self):
        if self._acme_account is None:
            self._acme_account = self.acme_client.acme_account

        return self._acme_account

    @property
    def acme_client(self):
        if self._acme_client is None:
            self._acme_client = AcmeClient(
                current_app.dns_client,
                acme_key=self._acme_key,
                acme_account=self._acme_account,
            )
        return self._acme_client

    lease_node_id = ConfigurationObject.basic_property("lease_node_id")

    @lease_node_id.setter  # type: ignore
    def lease_node_id(self, value):
        node_id = str(value) if value else None
        if self._lease_change_allowed(node_id):
            self._lease_node_id = node_id
        else:
            raise ValueError(
                "Lease is owned by {} until {}".format(
                    self.lease_node_id, self.lease_expiration
                )
            )

    lease_expiration = ConfigurationObject.basic_property("lease_expiration")

    def _lease_change_allowed(self, node_id=None):
        now = datetime.datetime.now(datetime.timezone.utc)
        if hasattr(self, "lease_node_id"):
            current_node = self.lease_node_id
        else:
            current_node = None

        if hasattr(self, "lease_expiration"):
            current_expiration = self.lease_expiration
        else:
            current_expiration = self.epoch
        # We want this to restrict changes to this so only one renew can be
        # in progress.
        #
        # Allow updates if
        #   -   current_expiration is unset or in the past
        #   -   the node matches the old node (either lease extension or
        #       termination)
        return (current_expiration is None or current_expiration < now) or (
            node_id == current_node
        )

    def __repr__(self):
        return "<AcmeAccount {}>".format(self.id)
