from typing import Optional, Union
from urllib.parse import urlparse

from flask import current_app

from globus.manager.api.models.connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway
from globus.manager.api.errors import InternalServerError
from globus.manager.constants import ConfigurationObjectTag


class BlackPearlLookupError(InternalServerError):
    error_description = (
        "Unable to read BlackPearl access file. This storage "
        "gateway's storage policies may be misconfigured."
    )


class BlackPearlStorageGateway(StorageGateway):
    CONNECTOR_ID = "7e3f3f5e-350c-4717-891a-2f451c24b0d4"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = ["s3_endpoint", "bp_access_id_file"]

    connector = Connector(id=CONNECTOR_ID, display_name="SpectraLogic BlackPearl")

    @property
    def s3_endpoint(self) -> str:
        value = self._s3_endpoint
        if value is None:
            raise AttributeError("s3_endpoint")

        return value

    @s3_endpoint.setter
    def s3_endpoint(self, value: str):
        self._s3_endpoint = value

    @checks_policy_property(prop="s3_endpoint", required=True)
    def check_s3_endpoint(self, value: str):
        if value is not None:
            try:
                parsed_value = urlparse(value)
                if parsed_value.scheme not in ["http", "https"]:
                    raise ValueError()
            except Exception:
                raise ValueError()
        elif value is None:
            raise ValueError()

    @property
    def bp_access_id_file(self) -> Union[bytes, str]:
        value = self._bp_access_id_file
        if value is None:
            raise AttributeError("bp_access_id_file")
        return value

    @bp_access_id_file.setter
    def bp_access_id_file(self, value: Optional[Union[bytes, str]]):
        self._bp_access_id_file = value

    @checks_policy_property(prop="bp_access_id_file", required=True)
    def check_bp_access_id_file(self, value: Union[bytes, str]):
        if value is not None and not isinstance(value, (bytes, str)):
            raise ValueError()

    def get_access_id(self, username):
        info = self.get_access_id_info(username)
        if info:
            return info[0]

    def get_secret_key(self, username):
        info = self.get_access_id_info(username)
        if info:
            return info[1]

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return user

    def check_user_allowed(self, username):
        if not super().check_user_allowed(username):
            return False

        return bool(self.get_access_id_info(username))

    def get_access_id_info(self, username):
        """
        Look up a given username in the BlackPearl
        access id file specified in the storage_policies and return a tuple
        containing the access_id and secret_key.
        """

        access_id_file = self.bp_access_id_file

        # error if the storage policies are misconfigured
        if not access_id_file:
            current_app.logger.error(
                "BlackPearl dsi config is missing bp_access_id_file."
            )
            raise BlackPearlLookupError()

        try:
            for line in open(access_id_file, "r"):
                vals = line.split()
                if len(vals) != 3:
                    continue
                if vals[0] == username:
                    return (vals[1], vals[2])
        except Exception as e:
            current_app.logger.error(
                "Error reading bp_access_id_file {}: {}".format(
                    self.bp_access_id_file, str(e)
                )
            )
            raise BlackPearlLookupError()

    @property
    def load_dsi_module(self):
        return "blackpearl"

    @load_dsi_module.setter
    def load_dsi_module(self, value: str):
        pass

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "blackpearl_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")
        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "blackpearl_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")
        super().patch_policies(**policies)

    def repr(self):
        return "<BlackPearlStorageGateway {}>".format(self.id)
