from urllib.parse import urlparse
import json
import socket

import boto
from boto.s3.connection import OrdinaryCallingFormat
from boto.exception import S3ResponseError
from cachelib import SimpleCache

from flask import current_app

from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.errors import GatewayTimeout
from .connector import Connector
from .storage_gateway import StorageGateway


user_info_cache = SimpleCache()


class CephConnectionError(GatewayTimeout):
    error_description = (
        "Unable to connect to Ceph. This storage gateway's "
        "storage policies may be misconfigured."
    )


class CephStorageGateway(StorageGateway):
    CONNECTOR_ID = "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = [
        "s3_endpoint",
        "s3_buckets",
        "ceph_admin_key_id",
        "ceph_admin_secret_key",
    ]

    connector = Connector(id=CONNECTOR_ID, display_name="Ceph")

    @property
    def s3_endpoint(self):
        value = self._s3_endpoint
        if value is None:
            raise AttributeError("s3_endpoint")

        return value

    @s3_endpoint.setter
    def s3_endpoint(self, value):
        if value is not None:
            try:
                parsed_value = urlparse(value)
                if parsed_value.scheme not in ["http", "https"]:
                    raise ValueError("Invalid s3_endpoint")
            except Exception:
                raise ValueError("Invalid s3_endpoint")
        elif value is None:
            raise ValueError("Invalid s3_endpoint")

        self._s3_endpoint = value
        self.configuration_hash.changed()

    @property
    def s3_buckets(self):
        value = self._s3_buckets
        if value is None:
            raise AttributeError("s3_buckets")

        return value

    @s3_buckets.setter
    def s3_buckets(self, value):
        if value is not None:
            if not isinstance(value, list):
                raise ValueError("Invalid s3_buckets")
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid s3_buckets")

        self._s3_buckets = value
        self.configuration_hash.changed()

    @s3_buckets.deleter
    def s3_buckets(self):
        self._s3_buckets = None
        self.configuration_hash.changed()

    @property
    def ceph_admin_key_id(self):
        value = self._ceph_admin_key_id
        if value is None:
            raise AttributeError("ceph_admin_key_id")
        return value

    @ceph_admin_key_id.setter
    def ceph_admin_key_id(self, value):
        if not value or not isinstance(value, str):
            raise ValueError("Invalid ceph_admin_key_id")
        self._ceph_admin_key_id = value
        self.configuration_hash.changed()

    @property
    def ceph_admin_secret_key(self):
        value = self._ceph_admin_secret_key
        if value is None:
            raise AttributeError("ceph_admin_secret_key")
        return value

    @ceph_admin_secret_key.setter
    def ceph_admin_secret_key(self, value):
        if not value or not isinstance(value, str):
            raise ValueError("Invalid ceph_admin_secret_key")
        self._ceph_admin_secret_key = value
        self.configuration_hash.changed()

    @property
    def load_dsi_module(self):
        return "ceph"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        if value is not None and value != self.load_dsi_module:
            raise ValueError("Invalid load_dsi_module")

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        if value is not None and value != self.process_user:
            raise ValueError("Invalid process_user")

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "ceph_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "ceph_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")
        super().patch_policies(**policies)

    def default_mapping(self, identity_data):
        current_app.logger.debug(
            "Attempting default mapping of %s with allowed_domains=%s",
            identity_data,
            self.allowed_domains,
        )

        if len(self.allowed_domains) == 1:
            username, domain = identity_data["username"].rsplit("@", 1)
            current_app.logger.debug("username=%s domain=%s", username, domain)
            if domain in self.allowed_domains:
                return username

    def get_user_display_name(self, username):
        user_info = self.__get_user_info(username)

        if user_info:
            return user_info["display_name"]

    def check_user_allowed(self, username):
        """
        Check if a given username maps to a credential set on the Ceph
        gateway described in the given storage_policies
        """
        # Check if the local username is allowed by policy
        if not super().check_user_allowed(username):
            return False

        # Check if the local username is has an account on storage system
        return bool(self.__get_user_info(username))

    def __get_user_info(self, username):
        user_info = user_info_cache.get(username)

        if user_info is None:
            s3_endpoint = self.s3_endpoint
            key_id = self.ceph_admin_key_id
            secret_key = self.ceph_admin_secret_key

            try:
                parsed = urlparse(s3_endpoint)

                host = parsed.hostname
                port = parsed.port
                use_ssl = parsed.scheme == "https"

                default_ports = {"http": 80, "https": 443}

                if port is None:
                    port = default_ports[parsed.scheme]
            except Exception as e:
                current_app.logger.error("Invalid s3_endpoint value: {}".format(str(e)))
                raise CephConnectionError()

            # clear any existing config to prevent conflicts, then set our own
            # values
            for section in boto.config.sections():
                boto.config.remove_section(section)
            boto.config.add_section("Boto")
            boto.config.set("Boto", "num_retries", "2")  # don't spin too long
            boto.config.add_section("s3")
            boto.config.set("s3", "use-sigv4", "True")

            # create the connection state for the admin bucket
            conn = boto.connect_s3(
                host=host,
                port=port,
                is_secure=use_ssl,
                aws_access_key_id=key_id,
                aws_secret_access_key=secret_key,
                calling_format=OrdinaryCallingFormat(),
            )

            conn.auth_region_name = "us-east-1"  # not actually used
            bucket = conn.get_bucket("/admin", validate=False)
            current_app.logger.debug(
                "About to connect to Ceph host {} "
                "on port {} and {}using ssl".format(
                    host, port, "not " if not use_ssl else ""
                )
            )

            # try to get info for the given username
            try:
                res = bucket.get_subresource("uid={}".format(username), "user")

                # sanity check that we got the expected user_id back
                json_body = json.loads(res)
                if json_body["user_id"] != username:
                    current_app.logger.error(
                        "Unexpected Ceph user id mismatch {} != {}".format(
                            username, json_body["user_id"]
                        )
                    )
                    return

                user_info = json_body

            # we expect to get a 404 for a username that does not map
            except S3ResponseError as e:
                if e.status == 404:
                    current_app.logger.debug("Ceph user {} not found".format(username))
                    return
                else:
                    current_app.logger.error(
                        "S3 Response Error on Ceph connection: {}".format(e)
                    )
                    raise CephConnectionError()
            except socket.error as e:
                current_app.logger.error(
                    "Socket error on Ceph connection: {}".format(e)
                )
                raise CephConnectionError()
            user_info_cache.set(username, user_info, 30)

        return user_info

    def __repr__(self):
        return "<CephStorageGateway {}>".format(self.display_name)
