import datetime
import os
import shutil

from flask import current_app
from sqlalchemy import orm
from OpenSSL import crypto

from .configuration_object import ConfigurationObject
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.schemas.appsync.certificate import certificate_schema
from globus.manager.utils.run_as_user import run_as_user


class Certificate(ConfigurationObject):
    _DATA_TYPE = "certificate#1.0.0"
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.CERTIFICATE.name}
    configuration_schema = certificate_schema

    def __init__(self, appsync_data=None, doc=None):
        self._x509_certificate = None

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.CERTIFICATE
            args["id"] = ConfigurationObjectTag.CERTIFICATE.name
            super().__init__(**args)
            self.init_on_load()
        elif doc is not None and (doc.get("DATA_TYPE") == self._DATA_TYPE):
            args = {k: v for k, v in doc.items()}
            super().__init__(
                tag=ConfigurationObjectTag.CERTIFICATE,
                id=ConfigurationObjectTag.CERTIFICATE.name,
            )
            for k, v in doc.items():
                if k != "DATA_TYPE":
                    setattr(self, k, v)
        else:
            raise ValueError("Unsupported DATA_TYPE")

    @orm.reconstructor
    def init_on_load(self):
        self._x509_certificate = None
        for k, v in self.configuration_hash.items():
            if k in ["domain_name", "certificate", "key", "chain"]:
                setattr(self, "_" + k, v)

    domain_name = ConfigurationObject.basic_property("domain_name")
    certificate = ConfigurationObject.basic_property("certificate")
    key = ConfigurationObject.basic_property("key")
    chain = ConfigurationObject.basic_property("chain")

    @property
    def expiration_time(self):
        """
        Expiration time as an zone-aware datetime object
        """
        cert = self.x509_certificate
        ts = cert.get_notAfter().decode()

        if ts.endswith("Z"):
            ts = ts[:-1] + "+0000"

        return datetime.datetime.strptime(ts, "%Y%m%d%H%M%S%z")

    @property
    def x509_certificate(self):
        """
        OpenSSL X.509 representation of this certificate
        """
        if not self._x509_certificate:
            self._x509_certificate = crypto.load_certificate(
                crypto.FILETYPE_PEM, self.certificate
            )
        return self._x509_certificate

    @staticmethod
    def create_certificate(domain):
        acme_account = ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        return Certificate(
            doc=dict(
                DATA_TYPE=Certificate._DATA_TYPE,
                **acme_account.get_cert_chain_and_key(domain)._asdict()
            )
        )

    def renew(self):
        acme_account = ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        res = acme_account.get_cert_chain_and_key(self.domain_name)
        self.certificate = res.certificate
        self.key = res.key
        self.chain = res.chain
        self.configuration_hash.changed()

    def _persist_to_filesystem(self, delete=False):
        cert_perms = 0o644
        cert_mode = os.O_CREAT | os.O_WRONLY
        key_perms = 0o600
        key_mode = os.O_CREAT | os.O_WRONLY

        domain_path = os.path.join(
            os.path.join(current_app.config["HOME"], "etc/letsencrypt/live"),
            self.domain_name,
        )

        if self.deleted:
            delete = True

        if delete:
            shutil.rmtree(domain_path, ignore_errors=True)
            return

        if not os.path.exists(domain_path):
            os.makedirs(domain_path, 0o700, exist_ok=True)

        def write_to_file(contents, path, perms, mode):
            with open(os.open(path, mode, perms), "w") as f:
                f.write(contents)

        write_to_file(
            self.certificate,
            os.path.join(domain_path, "cert.pem"),
            cert_perms,
            cert_mode,
        )

        write_to_file(
            self.chain, os.path.join(domain_path, "chain.pem"), cert_perms, cert_mode
        )

        write_to_file(
            "\n".join([self.certificate, self.chain]),
            os.path.join(domain_path, "fullchain.pem"),
            cert_perms,
            cert_mode,
        )

        write_to_file(
            self.key, os.path.join(domain_path, "privkey.pem"), key_perms, key_mode
        )

    @run_as_user(user="gcsweb")
    def write_config(self, delete=False):
        return self._persist_to_filesystem(delete=delete)

    def __repr__(self):
        return "<Certificate {}>".format(self.domain_name)
