import ipaddress
import time
import uuid

import dns.resolver
from flask import current_app
from sqlalchemy import orm

import globus_sdk

from .configuration import Configuration
from .configuration_object import ConfigurationObject
from globus.manager.constants import ConfigurationObjectTag, NodeStatus

import globus.manager.api.errors as errors
from globus.manager.api.schemas.appsync import endpoint_domain_schema


class Node:
    """
    Model for a GCS Node.

    Each Node object in the API consists of n ID, IP Address, and a status
    field.
    """

    _DATA_TYPE = "node#1.0.0"

    def __init__(
        self,
        DATA_TYPE=None,
        ip_addresses=None,
        id=None,
        status=None,
        incoming_port_range=None,
        outgoing_port_range=None,
    ):
        if DATA_TYPE != Node._DATA_TYPE:
            raise ValueError("Invalid DATA_TYPE for node")

        if id is None:
            self._id = str(uuid.uuid4())
        else:
            self._id = id

        self.ip_addresses = ip_addresses
        if not status:
            self.status = NodeStatus.active
        elif isinstance(status, NodeStatus):
            self.status = status
        else:
            try:
                self.status = getattr(NodeStatus, status)
            except AttributeError:
                raise ValueError("Invalid node status {}".format(status))
        if self.ip_addresses is None:
            raise ValueError("The ip_addresses property for a node may not be empty")

        self.incoming_port_range = incoming_port_range
        self.outgoing_port_range = outgoing_port_range

    @property
    def id(self):
        return self._id

    @property
    def ip_addresses(self):
        return self._ip_addresses

    @ip_addresses.setter
    def ip_addresses(self, value):
        if value is None:
            raise ValueError("Invalid ip_addresses")
        self._ip_addresses = [ipaddress.ip_address(ip) for ip in value]

    @property
    def incoming_port_range(self):
        if not self._incoming_port_range:
            raise AttributeError(
                "The incoming_port_range property is not defined for this node"
            )
        return self._incoming_port_range

    @staticmethod
    def _is_valid_port_range(start, end):
        return (
            start == int(start)
            and end == int(end)
            and start < end
            and start > 1
            and start < 2 ** 16 - 1
            and end > 1
            and end < 2 ** 16 - 1
        )

    @incoming_port_range.setter  # type: ignore
    def incoming_port_range(self, value):
        if (
            isinstance(value, (tuple, list))
            and len(value) == 2
            and self._is_valid_port_range(value[0], value[1])
        ):
            self._incoming_port_range = tuple(value)
        elif not value:
            self._incoming_port_range = None
        else:
            raise ValueError("Invalid incoming_port_range")

    @property
    def outgoing_port_range(self):
        if not self._outgoing_port_range:
            raise AttributeError(
                "The outgoing_port_range property is not defined for this node"
            )
        return self._outgoing_port_range

    @outgoing_port_range.setter  # type: ignore
    def outgoing_port_range(self, value):
        if (
            isinstance(value, (tuple, list))
            and len(value) == 2
            and self._is_valid_port_range(value[0], value[1])
        ):
            self._outgoing_port_range = tuple(value)
        elif not value:
            self._outgoing_port_range = None
        else:
            raise ValueError("Invalid outgoing_port_range")

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = NodeStatus(value)

    def put(self, node):
        changed = False
        if not isinstance(node, Node):
            if node.get("DATA_TYPE") != Node._DATA_TYPE:
                raise ValueError("Unknown DATA_TYPE")

            node_id_str = str(node["id"])
            ip_addresses = sorted(
                [ipaddress.ip_address(ip) for ip in node["ip_addresses"]]
            )
            status = NodeStatus(node["status"])
            incoming_port_range = node.get("incoming_port_range")
            outgoing_port_range = node.get("outgoing_port_range")
        else:
            node_id_str = str(node.id)
            ip_addresses = sorted(node.ip_addresses)
            status = NodeStatus(node.status)
            incoming_port_range = getattr(node, "incoming_port_range", None)
            outgoing_port_range = getattr(node, "outgoing_port_range", None)

        if node_id_str != self._id:
            raise ValueError("The id property is immutable")

        if ip_addresses != self._ip_addresses:
            self.ip_addresses = ip_addresses
            changed = True

        if status != self._status:
            self._status = status
            changed = True

        if incoming_port_range != self._incoming_port_range:
            self.incoming_port_range = incoming_port_range
            changed = True

        if outgoing_port_range != self._outgoing_port_range:
            self.outgoing_port_range = outgoing_port_range
            changed = True

        return changed

    def patch(self, node_changes):
        changed = False
        for n in node_changes:
            if n == "DATA_TYPE":
                if node_changes[n] != Node._DATA_TYPE:
                    raise ValueError("Unknown DATA_TYPE")
            elif n == "id":
                if str(node_changes[n]) != self.id:
                    raise ValueError("id is an immutable property")
            else:
                old_value = getattr(self, n)
                new_value = node_changes[n]
                if old_value != new_value:
                    if new_value is None:
                        delattr(self, n)
                    else:
                        setattr(self, n, new_value)
                    changed = True
        return changed

    def __eq__(self, other):
        return (
            self._id == other._id
            and self._ip_addresses == other._ip_addresses
            and self._status == other._status
            and self._incoming_port_range == other._incoming_port_range
            and self._outgoing_port_range == other._outgoing_port_range
        )

    def __repr__(self):
        return "<Node {}>".format(self.id)


class Domain(ConfigurationObject):
    __mapper_args__ = {
        "polymorphic_identity": ConfigurationObjectTag.WILDCARD_DOMAIN.name
    }
    _DATA_TYPE = endpoint_domain_schema._DATA_TYPE
    configuration_schema = endpoint_domain_schema

    def __init__(self, appsync_data=None, doc=None):
        self._nodes = {}
        self._domain_name = None

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["id"] = ConfigurationObjectTag.WILDCARD_DOMAIN.name
            args["tag"] = ConfigurationObjectTag.WILDCARD_DOMAIN
            super().__init__(**args)
            self.init_on_load()
        elif doc is not None and doc.get("DATA_TYPE") == self._DATA_TYPE:
            super().__init__(
                id=ConfigurationObjectTag.WILDCARD_DOMAIN.name,
                tag=ConfigurationObjectTag.WILDCARD_DOMAIN,
            )

            for k, v in doc.items():
                setattr(self, k, v)

            if self._domain_name is None:
                self.register_domain()
            elif self._domain_name is not None:
                self._check_status()
        else:
            raise ValueError("Unsupported DATA_TYPE")

    @orm.reconstructor
    def init_on_load(self):
        ch = self.configuration_hash
        if ch:
            self._nodes = {n["id"]: Node(**n) for n in ch.get("nodes", [])}
            self._domain_name = ch.get("domain_name")

    def _check_status(self):
        node_id = Configuration.query.filter(
            Configuration.name == "node_id"
        ).one_or_none()

        if node_id:
            node = self._nodes.get(node_id.value)
            if node is not None and node.status != current_app.node_status:
                current_app.node_status = node.status

    def register_domain(self):
        RETRIES = 15

        auth_client = current_app.auth_client
        dns_client = current_app.dns_client

        parsed_addresses = self.ip_addresses(active_only=True)

        a = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv4Address)]
        )
        aaaa = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv6Address)]
        )
        # Order so testing can assert more easily
        a.sort()
        aaaa.sort()

        domain_response = dns_client.create_domain(a=a, aaaa=aaaa)
        domain = domain_response.data["domain"]

        current_app.logger.debug("Waiting for domain propagation")
        for attempt in range(1, RETRIES):
            time.sleep(30)
            try:
                # Verify DNS has propagated
                current_app.logger.debug("Attempt {0}".format(attempt))
                dns.resolver.query("client_id." + domain, "TXT")
                break
            except dns.resolver.NXDOMAIN as e:
                if attempt + 1 < RETRIES:
                    current_app.logger.warning(
                        "DNS for {} not resolving, sleeping and trying again".format(
                            domain
                        )
                    )
                else:
                    raise e

            current_app.logger.debug(
                "DNS Propagated to us: registering domain with auth"
            )
            for attempt in range(1, RETRIES):
                try:
                    auth_response = auth_client.post(
                        "/v2/api/clients/{client_id}/fqdns".format(
                            client_id=auth_client.client_id
                        ),
                        json_body={"fqdn": domain},
                    )
                    current_app.logger.debug(
                        "Registered dns with auth returned {0}".format(
                            auth_response.http_status
                        )
                    )
                    assert auth_response.http_status == 201
                except globus_sdk.GlobusAPIError as e:
                    current_app.logger.debug(
                        "Error registering domain: {0} {1}".format(
                            e.http_status, str(e)
                        )
                    )
                    time.sleep(30)
                except Exception as e:
                    current_app.logger.debug("Error registering domain: {0}".format(e))
                    if attempt + 1 < RETRIES:
                        current_app.logger.debug(str(e))
                        time.sleep(30)
                    else:
                        raise e
        self._domain_name = domain
        return domain

    def ip_addresses(self, active_only=True):
        return [
            ip
            for n in self._nodes.values()
            for ip in n.ip_addresses
            if n.status == NodeStatus.active or not active_only
        ]

    def update_domain(self):
        if not self.domain_name:
            raise RuntimeError("No domain configured for update")

        dns_client = current_app.dns_client

        parsed_addresses = self.ip_addresses(active_only=True)
        a = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv4Address)]
        )
        aaaa = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv6Address)]
        )

        # Order so testing can assert more easily
        a.sort()
        aaaa.sort()

        dns_client.update_domain(self.domain_name, a=a, aaaa=aaaa)

    def delete(self):
        if self._domain_name:
            try:
                dns_client = current_app.dns_client
                dns_client.delete_domain(self._domain_name)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.error("delete_domain: {}".format(str(e)))
                raise errors.BadGatewayError("Error updating DNS")
            self._domain_name = None
            self._nodes = {}
        super().delete()

    def delete_domain(self, domain_name=None):
        if self._domain_name:
            if domain_name is None:
                domain_name = self._domain_name
            else:
                assert domain_name == self._domain_name
        else:
            assert domain_name is not None

        try:
            dns_client = current_app.dns_client
            dns_client.delete_domain(domain_name)
        except globus_sdk.GlobusAPIError as e:
            if e.http_status != 404:
                current_app.logger.error("delete_domain: {}".format(str(e)))
                raise errors.BadGatewayError("Error updating DNS")
        self._domain_name = None
        self._nodes = {}

    @property
    def nodes(self):
        if self._nodes:
            return list(self._nodes.values())
        else:
            return []

    @nodes.setter
    def nodes(self, value):
        nodes = {}
        if value:
            for n in value:
                if not isinstance(n, Node):
                    node = Node(**n)
                else:
                    node = n
                nodes[node.id] = node

        self._nodes = nodes

    @property
    def domain_name(self) -> str:
        return self._domain_name

    @domain_name.setter
    def domain_name(self, value: str):
        self._domain_name = value.lower()

    def add_node(self, node):
        if not isinstance(node, Node):
            node = Node(**node)

        if node.id in self._nodes:
            raise ValueError("Node already exists")

        ip_addresses = self.ip_addresses(active_only=False)
        for ip in node.ip_addresses:
            if ip in ip_addresses:
                raise ValueError("Node already exists")

        self._nodes[node.id] = node
        try:
            if self._domain_name:
                self.update_domain()
            else:
                self.register_domain()
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.error("add_node: {}: {}".format(node.id, str(e)))
            del self._nodes[node.id]
            raise errors.BadGatewayError("Error updating DNS")

    def get_node(self, id=None, ip_address=None):
        potential_nodes = [n for n in self._nodes.values()]
        if ip_address:
            potential_nodes = [
                n
                for n in potential_nodes
                if ipaddress.ip_address(ip_address) in n.ip_addresses
            ]

        if id:
            potential_nodes = [n for n in potential_nodes if n.id == id]

        if len(potential_nodes) > 1:
            raise ValueError("Multiple nodes match")

        elif len(potential_nodes) == 1:
            return potential_nodes.pop()

    def put_node(self, new_node):
        if not isinstance(new_node, Node):
            node = None
            node_id_str = str(new_node["id"])
            node = self._nodes.get(node_id_str)
        else:
            node_id_str = new_node.id
            node = self._nodes.get(new_node.id)

        if node is None:
            raise errors.NotFoundError("Unknown node {}".format(node_id_str))

        if node.put(new_node):
            if self._domain_name:
                self.update_domain()
            else:
                self.register_domain()
        self._check_status()

    def patch_node(self, node_changes):
        node_id_str = str(node_changes["id"])
        node = self._nodes.get(node_id_str)

        if node is None:
            raise errors.NotFoundError("Unknown node {}".format(node_id_str))

        if node.patch(node_changes):
            if self._domain_name:
                self.update_domain()
            else:
                self.register_domain()
        self._check_status()

    def remove_node(self, id=None, node=None, ip_address=None):
        potential_nodes = self._nodes.values()
        if id:
            potential_nodes = [n for n in potential_nodes if n.id == id]
        if node:
            potential_nodes = [n for n in potential_nodes if n.id == node.id]
        if ip_address:
            potential_nodes = [
                n
                for n in potential_nodes
                if ipaddress.ip_address(ip_address) in n.ip_addresses
            ]

        if len(potential_nodes) == 0:
            raise ValueError("Node does not exist")
        elif len(potential_nodes) > 1:
            raise ValueError("Multiple nodes match")
        else:
            old_node = potential_nodes[0]

        self._nodes.pop(old_node.id)

        if self.domain_name:
            self.update_domain()

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash,
        including updating DNS. This will fail if the domain name has been
        deleted.
        """
        current_domain_name = self._domain_name
        if data:
            self.update(data["items"][0])
        self.init_on_load()

        if not self._nodes:
            self.delete_domain(domain_name=current_domain_name)
        else:
            self.update_domain()

    def __repr__(self):
        return "<Domain {}>".format(self._domain_name)
