import random
from typing import Any, Dict, Optional, Union
from uuid import UUID

from flask import current_app
import marshmallow as ma
from jwcrypto import jwk
from sqlalchemy import orm

import globus_sdk
from .configuration_object import ConfigurationObject
from .decorators import DocumentChecker, checks_property
from .domain import Domain
import globus.manager.api.errors as errors
from globus.manager.constants import (
    ConfigurationObjectTag,
    NetworkUse,
)
from globus.manager.api.schemas.api.endpoint import endpoint_schema
from globus.manager.api.schemas.appsync.endpoint_config import endpoint_config_schema
from globus.manager.api.schemas.transfer import (
    non_functional_host_endpoint_create_schema,
    non_functional_host_endpoint_update_schema,
)
from globus.manager.constants import CollectionType
from globus.manager.api.utils.log import log_transfer_exception
from globus.manager.utils.obfuscate_int import obfuscate
from globus.manager.utils.platform_info import PlatformInfo


class Endpoint(ConfigurationObject, DocumentChecker):
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.ENDPOINT.name}
    _API_DATA_TYPE = "endpoint#1.0.0"
    _APPSYNC_DATA_TYPE = "endpoint_config#1.0.0"
    configuration_schema = endpoint_config_schema

    document_properties = [
        "public",
        "display_name",
        "subscription_id",
        "network_use",
        "max_concurrency",
        "preferred_concurrency",
        "max_parallelism",
        "preferred_parallelism",
        "organization",
        "department",
        "keywords",
        "description",
        "contact_email",
        "contact_info",
        "info_link",
    ]

    def __init__(self, appsync_data=None, doc=None):
        self._endpoint_id = None
        self._display_name = None
        self._oauth_state_key = None
        self._fingerprint = None

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.ENDPOINT
            super().__init__(**args)
            self.init_on_load()
        elif doc is not None and doc.get("DATA_TYPE") == self._API_DATA_TYPE:
            endpoint_id = doc.get("id", None)
            self.display_name = doc.get("display_name")
            self._network_use = NetworkUse(doc.get("network_use", "normal"))
            super().__init__(id=endpoint_id, tag=ConfigurationObjectTag.ENDPOINT)
            self.put(**doc)
        elif doc == {}:
            # Create a stub object at initialization time
            super().__init__(
                id=current_app.config["CLIENT_ID"], tag=ConfigurationObjectTag.ENDPOINT
            )
        else:
            raise ValueError("Unknown DATA_TYPE")

    @orm.reconstructor
    def init_on_load(self):
        if self.configuration_hash:
            ch = self.configuration_hash
            self._endpoint_id = ch.get("id")
            self._public = ch.get("public", True)
            self._display_name = ch.get("display_name")
            self._subscription_id = ch.get("subscription_id")
            self._network_use = NetworkUse(ch.get("network_use", "normal"))
            self._max_concurrency = ch.get("max_concurrency")
            self._preferred_concurrency = ch.get("preferred_concurrency")
            self._max_parallelism = ch.get("max_parallelism")
            self._preferred_parallelism = ch.get("preferred_parallelism")
            self._organization = ch.get("organization")
            self._department = ch.get("department")
            self._keywords = ch.get("keywords")
            self._description = ch.get("description")
            self._contact_email = ch.get("contact_email")
            self._contact_info = ch.get("contact_info")
            self._info_link = ch.get("info_link")
            self._domain_index = ch.get("domain_index", random.randint(1, 2 ** 16))
            self._oauth_state_key = ch.get("oauth_state_key")
            self._fingerprint = ch.get("fingerprint")

    public = ConfigurationObject.basic_property(
        "public",
        doc=(
            "Flag indicating whether this endpoint is visible to\n"
            "all other Globus users. If false, only users which have\n"
            "been granted a role on the endpoint or one of its collections,\n"
            "or belong to a domain allowed access to any of its storage\n"
            "gateways may view it."
        ),
    )
    display_name = ConfigurationObject.basic_property(
        "display_name",
        doc=(
            "Friendly name for the endpoint, not unique.  This may include\n"
            "unicode characters, but may not include new line characters."
        ),
    )
    oauth_state_key = ConfigurationObject.basic_property(
        "oauth_state_key", doc="Secret to encode and decode the OAuth state variable."
    )

    @property
    def subscription_id(self) -> Optional[str]:
        """
        This property is only present on a managed endpoint. If present, it is
        the string uuid of the subscription that the endpoint is managed by.
        """
        return self._subscription_id

    @subscription_id.setter
    def subscription_id(self, value: Union[str, UUID]):
        if isinstance(value, UUID):
            self._subscription_id = str(value)
        else:
            self._subscription_id = value
        current_app.subscription = self._subscription_id

    @property
    def collections(self):
        """
        Collection of collection objects which exist on this endpoint.
        """
        return ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.COLLECTION
        ).all()

    @property
    def storage_gateways(self):
        return ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.STORAGE_GATEWAY
        ).all()

    @property
    def endpoint_id(self) -> str:
        return self._endpoint_id

    @endpoint_id.setter
    def endpoint_id(self, value: Union[UUID, str]):
        if self._endpoint_id and self._endpoint_id != str(value):
            raise ValueError("The endpoint_id property is immutable")
        self._endpoint_id = str(value) if value else None

    @property
    def network_use(self) -> str:
        return getattr(self, "_network_use", NetworkUse.normal).name

    @network_use.setter
    def network_use(self, value: Union[NetworkUse, str]):
        if value:
            self._network_use = NetworkUse(value)
        else:
            self._network_use = NetworkUse.normal

        if self._network_use != NetworkUse.custom:
            self._max_concurrency = None
            self._preferred_concurrency = None
            self._max_parallelism = None
            self._preferred_parallelism = None

    @network_use.deleter
    def network_use(self):
        self.network_use = NetworkUse.normal

    @property
    def max_concurrency(self) -> int:
        if self._max_concurrency:
            return self._max_concurrency
        else:
            raise AttributeError(
                "The max_concurrency property is not defined for this endpoint"
            )

    @max_concurrency.setter
    def max_concurrency(self, value: int):
        if self.network_use != NetworkUse.custom.name:
            raise ValueError(
                "Can not set max_concurrency when network_use is {}".format(
                    self.network_use
                )
            )
        self._max_concurrency = int(value)

    @max_concurrency.deleter
    def max_concurrency(self):
        self._max_concurrency = None

    @property
    def preferred_concurrency(self) -> int:
        if self._preferred_concurrency:
            return self._preferred_concurrency
        else:
            raise AttributeError(
                "The preferred_concurrency property is not defined for this endpoint"
            )

    @preferred_concurrency.setter
    def preferred_concurrency(self, value: int):
        if self.network_use != NetworkUse.custom.name:
            raise ValueError(
                "Can not set preferred_concurrency when network_use is {}".format(
                    self.network_use
                )
            )
        self._preferred_concurrency = int(value)

    @preferred_concurrency.deleter
    def preferred_concurrency(self):
        self._preferred_concurrency = None

    @property
    def max_parallelism(self) -> int:
        if self._max_parallelism:
            return self._max_parallelism
        else:
            raise AttributeError(
                "The max_parallelism property is not defined for this endpoint"
            )

    @max_parallelism.setter
    def max_parallelism(self, value: int):
        if self.network_use != NetworkUse.custom.name:
            raise ValueError(
                "Can not set max_parallelism when network_use is {}".format(
                    self.network_use
                )
            )
        self._max_parallelism = int(value)

    @max_parallelism.deleter
    def max_parallelism(self):
        self._max_parallelism = None

    @property
    def preferred_parallelism(self) -> int:
        if self._preferred_parallelism:
            return self._preferred_parallelism
        else:
            raise AttributeError(
                "The preferred_parallelism property is not defined for this endpoint"
            )

    @preferred_parallelism.setter
    def preferred_parallelism(self, value: int):
        if self.network_use != NetworkUse.custom.name:
            raise ValueError(
                "Can not set preferred_parallelism when network_use is {}".format(
                    self.network_use
                )
            )
        self._preferred_parallelism = int(value)

    @preferred_parallelism.deleter
    def preferred_parallelism(self):
        self._preferred_parallelism = None

    organization = ConfigurationObject.basic_property("organization")
    department = ConfigurationObject.basic_property("department")
    keywords = ConfigurationObject.basic_property("keywords")
    description = ConfigurationObject.basic_property("description")
    contact_email = ConfigurationObject.basic_property("contact_email")
    contact_info = ConfigurationObject.basic_property("contact_info")
    info_link = ConfigurationObject.basic_property("info_link")

    @property
    def allow_udt(self) -> bool:
        """
        Allow data transfer on this endpoint using the UDT protocol. This
        feature is currently not implemented, so this value is always False.
        """
        return False

    @allow_udt.setter
    def allow_udt(self, value: bool):
        pass

    @checks_property(prop="allow_udt")
    def check_allow_udt(self, value: bool):
        if value:
            raise ValueError(
                "The allow_udt property may not be set to true on this endpoint."
            )

    @property
    def gcs_manager_url(self) -> str:
        """
        URL of the GCS Manager API service for this endpoint.
        """
        try:
            domain = Domain.query.one().domain_name
            return "https://{}".format(domain)
        except orm.exc.NoResultFound:
            raise AttributeError(
                "The gcs_manager_url property is not defined for this endpoint"
            )

    # Index value used to create obfuscated domains
    domain_index = ConfigurationObject.basic_property("domain_index")

    def allocate_domain_name(self, collection_type: str = "guest") -> str:
        """
        Allocate the next pseudo-random domain name for a new collection.
        This updates the `domain_index` property of this endpoint, so this
        endpoint object must be committed to appsync prior to using this
        domain name.

        :param str collection_type:
            A string containing either `guest` or `mapped` to declare the
            collection type.

        :return:
            Returns a new pseudo-random domain name for this collection
        :rtype: str

        :raises ValueError:
            The `collection_type` value is not one of the allowed values
        """
        ctype = CollectionType(collection_type).name
        self.domain_index += 1
        domain_name = obfuscate(self.domain_index)
        if ctype == "guest":
            prefix = "g-"
        else:
            prefix = "m-"

        endpoint_domain = Domain.query.one().domain_name

        full_domain_name = "{}{}.{}".format(prefix, domain_name, endpoint_domain)
        self.refresh_configuration_hash()
        return full_domain_name

    def _update_endpoint_in_globus_services(
        self, create=True, changed_display_name=False
    ):
        """
        Internal wrapper for pushing updates to "external" Globus services,
        namely: Transfer and Search

        Returns a tuple
            (TransferResponse, SearchResponse)

        Whenever an endpoint is created or updated, this should be run in order
        to sync those changes to those services as appropriate.
        """
        # update Transfer
        if create:
            try:
                transfer_response = current_app.transfer_client.create_endpoint(
                    non_functional_host_endpoint_create_schema.dump(self)
                )
            except globus_sdk.GlobusAPIError as e:
                log_transfer_exception(e)
                raise
        else:
            try:
                transfer_response = current_app.transfer_client.update_endpoint(
                    self.endpoint_id,
                    non_functional_host_endpoint_update_schema.dump(self),
                )
            except globus_sdk.GlobusAPIError as e:
                log_transfer_exception(e)
                if e.http_status != 404:
                    raise errors.BadGatewayError("Error updating endpoint")

        if changed_display_name or create:
            try:
                self._update_scopes()
            except Exception as e:
                current_app.logger.warning(
                    "Unable to update scope, name and description "
                    "may not reflect the current endpoint display_name: %s %s",
                    type(e),
                    str(e),
                )

        # TODO: Integration with subscription service
        # TODO: update Search
        search_response = None
        return (transfer_response, search_response)

    def _update_scopes(self):
        manage_collections_scope_string = "urn:globus:auth:scope:{}:manage_collections".format(
            self.id
        )
        scopes = current_app.auth_client.get(
            "/v2/api/scopes", params={"scope_strings": manage_collections_scope_string}
        )
        manage_collections_scope = None

        for s in scopes.data["scopes"]:
            if s["client"] == self.id:
                manage_collections_scope = s
                dependent_scope_strings = [
                    "openid",
                    "profile",
                    "email",
                    "urn:globus:auth:scope:auth.globus.org:view_identities",
                    "urn:globus:auth:scope:transfer.api.globus.org:all",
                    "urn:globus:auth:scope:groups.api.globus.org:"
                    "view_my_groups_and_memberships",
                ]
                dependent_scopes = current_app.auth_client.get(
                    "/v2/api/scopes",
                    params={"scope_strings": ",".join(dependent_scope_strings)},
                ).data
                assert len(dependent_scopes["scopes"]) == len(dependent_scope_strings)
                display_name = "Manage collections on {}".format(self.display_name)
                description = (
                    "Access to all capabilities of the Globus Connect Server. "
                    "This includes creation, deletion and management of "
                    "collections and storage system credentials on the server "
                    "{}".format(self.display_name)
                )

                current_app.auth_client.put(
                    "/v2/api/scopes/{}".format(manage_collections_scope["id"]),
                    {
                        "scope": {
                            "name": display_name,
                            "description": description,
                            "dependent_scopes": [
                                {
                                    "scope": s["id"],
                                    "optional": False,
                                    "requires_refresh_token": False,
                                    "allows_refresh_token": True,
                                }
                                for s in dependent_scopes["scopes"]
                            ],
                        }
                    },
                )

    @property
    def configuration(self):
        if self._endpoint_id:
            return ConfigurationObject.configuration.fget(self)
        else:
            return "{}"

    @configuration.setter
    def configuration(self, value):
        return ConfigurationObject.configuration.fset(self, value)

    @property
    def fingerprint(self) -> Dict[str, Any]:
        """
        System fingprint. This value is a JSON blob containing at least the
        GCS version.
        """
        fp = self._fingerprint
        if fp:
            return fp
        raise AttributeError("fingerprint")

    @fingerprint.setter
    def fingerprint(self, value):
        if self._fingerprint != value:
            self._fingerprint = value

    def update_fingerprint(
        self, new_fingerprint: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Update the system fingerprint for this endpoint. This updates the
        transfer document for this collection if the fingerprint has changed
        from what was in the database.

        :param new_fingerprint:
            Optional new fingerprint value. If not supplied, this function will
            compute the fingerprint.

        :returns bool:
            Returns True if the fingerprint changed and was updated in
            transfer; False otherwise.
        """
        old_fingerprint = getattr(self, "_fingerprint", None)

        if new_fingerprint is None:
            new_fingerprint = PlatformInfo().fingerprint()

        if new_fingerprint != old_fingerprint:
            self.fingerprint = new_fingerprint
            self._update_endpoint_in_globus_services(create=False)
            return True
        return False

    def delete(self):
        endpoint_id = self.endpoint_id

        if endpoint_id is not None:
            try:
                res = current_app.transfer_client.delete_endpoint(endpoint_id)
                self._endpoint_id = None
                return res
            except globus_sdk.exc.GlobusAPIError as e:
                log_transfer_exception(e)
                if e.http_status != 404:
                    raise
                self._endpoint_id = None

    def _check_network_parameters(self):
        if self._network_use != NetworkUse.normal and not self.subscription_id:
            raise errors.SubscriptionRequiredError(
                "The network_use may only be changed on a managed endpoint",
                subscription_level="starter",
            )

        if self.network_use == NetworkUse.custom.name:
            missing_params = []
            network_params = [
                "max_concurrency",
                "max_parallelism",
                "preferred_concurrency",
                "preferred_parallelism",
            ]
            for param in network_params:
                if getattr(self, param, None) is None:
                    missing_params.append(param)

            if missing_params:
                if len(missing_params) == 1:
                    missing_message = missing_params[0]
                elif len(missing_params) == 2:
                    missing_message = " and ".join(missing_params)
                else:
                    missing_message = (
                        ", ".join(missing_params[:-1]) + ", and " + missing_params[-1]
                    )
                raise errors.ValidationError(
                    "Custom network_use requires settings for {}".format(
                        missing_message
                    ),
                    exc=ma.ValidationError(
                        {mp: "Missing data for required field" for mp in missing_params}
                    ),
                )

    def patch(self, DATA_TYPE=None, **kwargs):
        if DATA_TYPE != self._API_DATA_TYPE:
            raise errors.ValidationError("Unsupported DATA_TYPE")

        old_display_name = getattr(self, "display_name", None)

        self.check_patch_doc(**kwargs)
        for p in self.document_properties:
            if p in kwargs:
                setattr(self, p, kwargs[p])

        self._check_network_parameters()
        changed_display_name = old_display_name != self.display_name

        self._update_endpoint_in_globus_services(
            create=False, changed_display_name=changed_display_name
        )

        self.refresh_configuration_hash()

    def put(self, DATA_TYPE=None, **kwargs):
        if DATA_TYPE != self._API_DATA_TYPE:
            raise errors.ValidationError("Unsupported DATA_TYPE")

        old_display_name = getattr(self, "display_name", None)

        self.check_doc(**kwargs)
        for p in self.document_properties:
            if not endpoint_schema.fields[p].dump_only:
                if p not in kwargs:
                    if hasattr(self, p):
                        setattr(self, p, None)
                else:
                    setattr(self, p, kwargs[p])

        self._check_network_parameters()
        changed_display_name = old_display_name != self.display_name

        if not hasattr(self, "domain_index"):
            self.domain_index = random.randint(1, 2 ** 16)

        if self._oauth_state_key is None:
            self._oauth_state_key = jwk.JWK.generate(kty="oct", size=256).get_op_key()

        if not self._endpoint_id:
            res, _ = self._update_endpoint_in_globus_services()
            self._endpoint_id = res["id"]
        else:
            self._update_endpoint_in_globus_services(
                create=False, changed_display_name=changed_display_name
            )
        self.refresh_configuration_hash()

    def rollback(self, data=None):
        if data:
            self.update(data["items"][0])
        else:
            self.init_on_load()
        self._update_endpoint_in_globus_services(
            create=False, changed_display_name=True
        )
        self.refresh_configuration_hash()

    def __repr__(self):
        return "<Endpoint {}>".format(self._display_name)
