import datetime
import httplib2
from flask import current_app

import globus.manager.api.errors as errors
from oauth2client import client as google_oauth
from oauth2client import GOOGLE_TOKEN_URI

from .user_credential import UserCredential
from .decorators import checks_property


class GoogleCloudPlatformUserCredential(UserCredential):
    epoch = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)

    connector_properties = [
        "sub",
        "email",
        "access_token",
        "refresh_token",
        "token_expiry",
        "scopes",
    ]

    @property
    def sub(self):
        value = self._sub
        if not value:
            raise AttributeError("sub")
        return value

    @sub.setter
    def sub(self, value):
        self._sub = value

    @sub.deleter
    def sub(self):
        self._sub = None

    @property
    def email(self):
        value = self._email
        if not value:
            raise AttributeError("email")
        return value

    @email.setter
    def email(self, value):
        self._email = value

    @email.deleter
    def email(self):
        self._email = None

    @property
    def access_token(self):
        value = self._access_token
        if not value:
            raise AttributeError("access_token")
        return value

    @access_token.setter
    def access_token(self, value):
        self._access_token = value

    @access_token.deleter
    def access_token(self):
        self._access_token = None

    @property
    def refresh_token(self):
        value = self._refresh_token
        if not value:
            raise AttributeError("refresh_token")
        return value

    @refresh_token.setter
    def refresh_token(self, value):
        self._refresh_token = value
        if not value:
            self.invalid = True

    @refresh_token.deleter
    def refresh_token(self):
        self._refresh_token = None
        self.invalid = True

    @property
    def token_expiry(self):
        value = self._token_expiry
        if not value:
            raise AttributeError("token_expiry")
        return value

    @token_expiry.setter
    def token_expiry(self, value):
        if value.tzinfo is None:
            self._token_expiry = value.replace(tzinfo=datetime.timezone.utc)
        else:
            self._token_expiry = value

    @token_expiry.deleter
    def token_expiry(self):
        self._token_expiry = self.epoch

    @property
    def scopes(self):
        value = self._scopes
        if not value:
            raise AttributeError("scopes")
        return value

    @scopes.setter
    def scopes(self, value):
        self._scopes = value

    @scopes.deleter
    def scopes(self):
        self._scopes = None

    def needs_refresh(self):
        """
        Return True if credential requires some refresh action to be performed
        before the credential can be used. Subclasses should implement this
        if necessary to check for expiring credentials.
        """
        if self.deleted:
            return False

        try:
            five_mins_before_expiration = self.token_expiry - datetime.timedelta(
                minutes=5
            )
            now = datetime.datetime.now(datetime.timezone.utc)

            return self.invalid or (now > five_mins_before_expiration)
        except AttributeError:
            return True

    def refresh(self):
        """
        Update an expired temporary credential. Subclasses should implement
        this if necessary to refresh credentials.
        """
        try:
            gcred = google_oauth.GoogleCredentials(
                access_token=None,
                client_id=self.storage_gateway.client_id,
                client_secret=self.storage_gateway.secret,
                refresh_token=self.refresh_token,
                token_expiry=None,
                token_uri=GOOGLE_TOKEN_URI,
                user_agent=None,
            )
            gcred.refresh(httplib2.Http())
        except Exception as e:
            current_app.logger.warning("Google Credential refresh failure: %s", str(e))
            raise errors.InvalidCredentialError(user_credential_id=self.id)

        self.refresh_token = gcred.refresh_token
        self.access_token = gcred.access_token
        self.token_expiry = gcred.token_expiry

        self.configuration_hash.changed()

    def delete(self):
        try:
            token = self.refresh_token
        except AttributeError:
            token = None
        if token:
            gcred = google_oauth.GoogleCredentials(
                access_token=None,
                client_id=None,
                client_secret=None,
                refresh_token=token,
                token_expiry=None,
                token_uri=GOOGLE_TOKEN_URI,
                user_agent=None,
            )
            try:
                gcred.revoke(httplib2.Http())
            except google_oauth.TokenRevokeError as e:
                current_app.logger.debug("delete: revoke: %s: %s", self.id, str(e))
            finally:
                self.refresh_token = None
                self.access_token = None
        super().delete()

    @checks_property(prop="display_name")
    def check_display_name(self, value: str):
        if value and getattr(self, "display_name", None) != value:
            raise AttributeError()

    def patch(self, **doc):
        display_name = doc.pop("display_name", getattr(self, "display_name", None))
        doc["display_name"] = display_name
        policies = doc.get("policies", None)
        if policies:
            if policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
                raise ValueError("Unsupported DATA_TYPE")
            for p in policies:
                if p != "DATA_TYPE" and getattr(self, p, None) != policies[p]:
                    raise ValueError("Cannot change {} via PATCH".format(p))

        if getattr(self, "refresh_token", None) is None:
            if not doc.get("invalid", True):
                raise ValueError("Cannot mark this credential as valid")
            doc["invalid"] = True

        super().patch(**doc)

    def put(self, **doc):
        # Users cannot change policies via put, must be through put_credentials,
        # since those come from the google tokens.
        display_name = doc.pop("display_name", getattr(self, "display_name", None))
        doc["display_name"] = display_name

        policies = doc.get("policies", None)

        if policies:
            if policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
                raise ValueError("Unsupported DATA_TYPE")
            for p in policies:
                if p != "DATA_TYPE" and getattr(self, p, None) != policies[p]:
                    raise ValueError("Cannot change {} via PUT".format(p))

        if getattr(self, "refresh_token", None) is None:
            if not doc.get("invalid", True):
                raise ValueError("Cannot mark this credential as valid")
            doc["invalid"] = True

        super().put(**doc)

    def put_credentials(self, credentials):
        patch = self._credentials_to_patch(credentials)
        self._patch(**patch)
        self.refresh_configuration_hash()

    def _credentials_to_patch(self, credentials):
        if credentials:
            id_token = credentials.id_token
            refresh_token = credentials.refresh_token
            access_token = credentials.access_token

            patch = {
                "DATA_TYPE": self._DATA_TYPE,
                "display_name": id_token["name"],
                "policies": {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "sub": id_token["sub"],
                    "refresh_token": refresh_token,
                    "email": id_token["email"],
                    "access_token": access_token,
                    "token_expiry": credentials.token_expiry,
                    "scopes": credentials.scopes,
                },
                "invalid": False,
                "deleted": False,
            }
        else:
            patch = {
                "display_name": "{} (unprovisioned)".format(self.username),
                "policies": {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_token": None,
                    "refresh_token": None,
                    "token_expiry": self.epoch,
                },
                "deleted": False,
            }
        patch["invalid"] = patch["policies"]["refresh_token"] is None
        if not patch["invalid"]:
            patch["provisioned"] = True

        return patch
