from collections.abc import Sequence
from typing import List, Optional

from flask import current_app

from .connector import Connector
from .decorators import checks_policy_property
from .domain import Domain
from .storage_gateway import StorageGateway

from globus.manager.constants import ConfigurationObjectTag


class GoogleCloudStorageGateway(StorageGateway):
    CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = [
        "client_id",
        "secret",
        "buckets",
        "projects",
    ]

    connector = Connector(
        id=CONNECTOR_ID, display_name="Google Cloud Storage", can_post_credential=False
    )

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value: str):
        self._client_id = value

    @checks_policy_property(prop="client_id", required=True)
    def check_client_id(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Invalid client_id")

    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self, value):
        self._secret = value

    @checks_policy_property(prop="secret", required=True)
    def checks_secret(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Invalid secret")

    @property
    def buckets(self):
        value = self._buckets
        if value is None:
            raise AttributeError("buckets")
        return value

    @buckets.setter
    def buckets(self, value):
        self._buckets = value

    @buckets.deleter
    def buckets(self):
        self._buckets = None

    @checks_policy_property(prop="buckets")
    def check_buckets(self, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    @property
    def projects(self) -> List[str]:
        value = self._projects
        if value is None:
            raise AttributeError("projects")
        return value

    @projects.setter
    def projects(self, value):
        self._projects = value

    @projects.deleter
    def projects(self):
        self._projects = None

    @checks_policy_property(prop="projects")
    def check_projects(self, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError("Invalid projects")
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid projects")

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    def check_user_allowed(self, username):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    @property
    def load_dsi_module(self) -> str:
        return "google_cloud"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        pass

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """
        endpoint_domain = Domain.query.one().domain_name
        return "https://{}/api/v1/authcallback_google".format(endpoint_domain)

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_cloud_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_cloud_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().patch_policies(**policies)
