import httplib2
import json

import oauth2client
from oauth2client.client import HttpAccessTokenRefreshError
from flask import current_app

import globus.manager.api.errors as errors
from globus.manager.utils.update_query_string import update_query_string
from .google_cloud_platform_user_credential import GoogleCloudPlatformUserCredential
from globus.manager.constants import ConfigurationObjectTag


class GoogleCloudStorageUserCredential(GoogleCloudPlatformUserCredential):
    POLICY_DATA_TYPE = "google_cloud_storage_user_credential_policies#1.0.0"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name
            + "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
        )
    }

    def __init__(
        self, appsync_data=None, doc=None, storage_gateway=None, credentials=None
    ):
        if doc:
            self.username = doc["username"]
            new_doc = {**doc}
            new_doc.pop("display_name", None)
            # Policies must be either empty (for an unprovisioned credential)
            # or passed in as a credentials object from the google callback
            policies = new_doc.get("policies", None)
            if policies is not None and policies != {
                "DATA_TYPE": self.POLICY_DATA_TYPE
            }:
                raise ValueError("policies")
            patch = self._credentials_to_patch(credentials)
            patch["deleted"] = False
            patch["provisioned"] = bool(credentials)

            doc = {**new_doc, **patch}

        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

    @property
    def connector_id(self):
        return "56366b96-ac98-11e9-abac-9cb6d0d9fd63"

    @connector_id.setter
    def connector_id(self, value):
        pass

    def can_access_project(self, project):
        """
        Confirm that the project is viewable by this user credential
        """
        # confirm project is accessible by user_credential
        refresh_token = self.refresh_token
        secret = self.storage_gateway.secret
        client_id = self.storage_gateway.client_id
        cred = oauth2client.client.GoogleCredentials(
            None,
            client_id,
            secret,
            refresh_token,
            None,
            "https://accounts.google.com/o/oauth2/token",
            "globus-connect-server-manager",
        )
        http = cred.authorize(httplib2.Http())
        url = "https://cloudresourcemanager.googleapis.com/v1/projects/{}"
        response, content = http.request(url.format(project))

        if int(response.get("status", 0)) != 200:
            raise errors.ForbiddenError(
                "Your credentials do not have permissions to access the Google project {}".format(
                    project,
                )
            )
        else:
            # sanity check we got back the expected project
            data = json.loads(content.decode("utf8"))
            if data.get("projectId") != project:
                raise errors.InternalServerError("Unexpected Google Cloud response")
        return True

    @property
    def projects(self):
        if not hasattr(self, "_projects") or self._projects is None:
            if self.invalid:
                raise AttributeError(
                    "The projects property is not defined for this credential"
                )

            current_app.logger.debug("Looking up projects")
            refresh_token = self.refresh_token
            secret = self.storage_gateway.secret
            client_id = self.storage_gateway.client_id

            try:
                cred = oauth2client.client.GoogleCredentials(
                    None,
                    client_id,
                    secret,
                    refresh_token,
                    None,
                    "https://accounts.google.com/o/oauth2/token",
                    "globus-connect-server-manager",
                )
                http = cred.authorize(httplib2.Http())
            except Exception as e:
                current_app.logger.warning("invalid credentials: %s", str(e))
                self.invalid = True
                raise AttributeError(
                    "The projects property is not defined for this credential"
                )

            final_page = False
            page_token = None
            projects = []
            while not final_page:
                url = "https://cloudresourcemanager.googleapis.com/v1/projects"
                if page_token:
                    url = update_query_string(
                        url,
                        dict(
                            page_token=page_token,
                            fields="nextPageToken,projects(projectId,name))",
                        ),
                    )
                try:
                    response, content = http.request(url)
                    status = int(response.get("status", 0))
                except HttpAccessTokenRefreshError as e:
                    current_app.logger.warning("invalid credentials: %s", str(e))
                    self.invalid = True
                    raise AttributeError(
                        "The projects property is not defined for this credential"
                    )
                except Exception as e:
                    current_app.logger.warning("project enumeration error: %s", str(e))
                    raise AttributeError(
                        "The projects property is not defined for this credential"
                    )

                if status == 200:
                    data = json.loads(content.decode("utf8"))
                    page_token = data.get("nextPageToken")
                    final_page = page_token is None
                    projects.extend(data.get("projects", []))
                # if we get back a 4xx error, the credential is no longer valid
                elif status == 401:
                    self.invalid = True
                    projects = []
                    break
                # other errors could be temporary so we don't invalidate the
                # cred
                else:
                    projects = []
                    break

            # If google_cloud_storage_projects in the Storage Gateway config,
            # filter out any projects that are not in that list
            try:
                allowed_projects = self.storage_gateway.projects
            except AttributeError:
                allowed_projects = []
            if allowed_projects:
                projects = list(
                    [p for p in projects if p["projectId"] in allowed_projects]
                )
            self._projects = projects

        return [
            {"projectId": p["projectId"], "name": p["name"]} for p in self._projects
        ]

    @projects.deleter
    def projects(self):
        self._projects = None

    def __repr__(self):
        return "<GoogleCloudStorageUserCredential {}>".format(self.id)
