from cachelib import SimpleCache
import requests
from flask import current_app

from .collection import Collection
from globus.manager.constants import ConfigurationObjectTag
import globus.manager.api.errors as errors


class GoogleDriveCollection(Collection):
    POLICIES_DATA_TYPE = "google_drive_collection_policies#1.0.0"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "976cf0cf-78c3-4aab-82d2-7c16adbcc281"
        )
    }
    cache = SimpleCache()

    def check_user_credential(self, credential):
        """
        Check whether the given credential can be used to access this
        collection. To do this, we'll hit the google drive about endpoint
        which gives us info about the user's ability to access drive. This
        may mark the credential as invalid, but will not commit the change.

        The result is cached for 30 seconds.
        """
        cache_key = credential.access_token
        about = self.cache.get(cache_key)

        current_app.logger.debug("Cached result for %s is %s", credential, about)
        if about is None:
            resource = "https://www.googleapis.com/drive/v3/about"
            params = {"fields": "kind"}
            headers = {"Authorization": "Bearer {}".format(credential.access_token)}
            response = requests.get(resource, params=params, headers=headers)

            current_app.logger.debug(
                "Credential validity check: %s", response.status_code
            )
            if response.status_code >= 200 and response.status_code < 400:
                about = True
            else:
                about = errors.InvalidCredentialError(
                    "Your credentials do not have sufficient access to access this collection",
                    user_credential_id=credential.id,
                )
                credential.invalid = True
            self.cache.set(cache_key, about, 30)

        if isinstance(about, Exception):
            raise about
        elif about:
            return True

    @property
    def connector_id(self):
        return "976cf0cf-78c3-4aab-82d2-7c16adbcc281"

    @connector_id.setter
    def connector_id(self, value):
        pass

    def __repr__(self):
        return "<GoogleDriveCollection {}>".format(self.id)
