from flask import current_app

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway

from globus.manager.constants import ConfigurationObjectTag


class GoogleDriveStorageGateway(StorageGateway):
    CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = [
        "client_id",
        "secret",
        "user_api_rate_quota",
    ]

    connector = Connector(
        id=CONNECTOR_ID, display_name="Google Drive", can_post_credential=False
    )

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        self._client_id = value

    @checks_policy_property(prop="client_id", required=True)
    def check_client_id(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError()

    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self, value):
        self._secret = value

    @checks_policy_property(prop="secret", required=True)
    def check_secret(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError()

    @property
    def user_api_rate_quota(self):
        if self._user_api_rate_quota is None:
            raise AttributeError(
                "The user_api_rate_quota property is not defined for this storage gateway"
            )

        return self._user_api_rate_quota

    @user_api_rate_quota.setter
    def user_api_rate_quota(self, value):
        self._user_api_rate_quota = value
        self.configuration_hash.changed()

    @user_api_rate_quota.deleter
    def user_api_rate_quota(self):
        self._user_api_rate_quota = None

    @checks_policy_property(prop="user_api_rate_quota")
    def check_user_api_rate_quota(self, value: int):
        if value is not None and not isinstance(value, int):
            raise ValueError("Invalid user_api_rate_quota")

    def check_user_allowed(self, username):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    @property
    def load_dsi_module(self):
        return "google_drive"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        if value is not None and value != self.load_dsi_module:
            raise ValueError("Invalid load_dsi_module")

    def home_directory(self, username):
        return "/My Drive"

    def delete(self):
        self._client_id = None
        self._secret = None
        super().delete()

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """
        endpoint_domain = current_app.config["SERVER_NAME"]
        return "https://{}/api/v1/authcallback_google".format(endpoint_domain)

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        if value is not None and value != self.process_user:
            raise ValueError("Invalid process_user")

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_drive_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_drive_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")
        super().patch_policies(**policies)
