from collections.abc import Sequence
import ctypes
import grp
import json
import pwd
from typing import List, Optional

from flask import current_app

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway
from globus.manager.constants import ConfigurationObjectTag

_utils = None

try:
    _utils = ctypes.CDLL("libglobus_connect_gridftp_server_utils.so")
    _utils.globus_gfs_check_user_allowed_json.restype = ctypes.c_bool
    _utils.globus_gfs_check_user_allowed_json.argtypes = (
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_int),
        ctypes.c_int,
        ctypes.c_char_p,
    )
except OSError:
    pass


class PosixStorageGateway(StorageGateway):
    CONNECTOR_ID = "145812c8-decc-41f1-83cf-bb2a85a2a70b"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector_policy_names = ["groups_allow", "groups_deny"]

    connector = Connector(id=CONNECTOR_ID, display_name="POSIX")

    @staticmethod
    def __validate_string_list(name, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError("Invalid value for the {} property.".format(name))
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid value for the {} property.".format(name))

    @property
    def groups_allow(self):
        value = self._groups_allow
        if value is None:
            raise AttributeError(
                "The groups_allow property is not set on this storage gateway."
            )
        return value

    @groups_allow.setter
    def groups_allow(self, value):
        self._groups_allow = value

    @groups_allow.deleter
    def groups_allow(self):
        self._groups_allow = None

    @checks_policy_property(prop="groups_allow")
    def _check_groups_allow(self, value: Optional[List[str]]) -> None:
        PosixStorageGateway.__validate_string_list("groups_allow", value)

    @property
    def groups_deny(self):
        value = self._groups_deny
        if value is None:
            raise AttributeError(
                "The groups_deny property is not set on this storage gateway."
            )
        return value

    @groups_deny.setter
    def groups_deny(self, value):
        self._groups_deny = value

    @groups_deny.deleter
    def groups_deny(self):
        self._groups_deny = None

    @checks_policy_property(prop="groups_deny")
    def _check_groups_deny(self, value: Optional[List[str]]):
        PosixStorageGateway.__validate_string_list("groups_deny", value)

    def __repr__(self):
        return "<PosixStorageGateway {}>".format(self.display_name)

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "posix_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "posix_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().patch_policies(**policies)

    def home_directory(self, username):
        pwd, _ = self.lookup_unix_info(username, include_groups=False)

        if pwd:
            return pwd.pw_dir

    def get_user_display_name(self, username):
        pwd, _ = self.lookup_unix_info(username, include_groups=False)

        if pwd:
            return pwd.pw_gecos.split(",")[0] or username
        else:
            return username

    def lookup_unix_info(self, username, include_groups=True):
        """
        Helper method to resolve the given username to a passwd entry and
        supplement group list
        """
        supplemental_group_ids = None

        try:
            passwd_entry = pwd.getpwnam(username)
        except KeyError:
            passwd_entry = None

        if passwd_entry is not None and include_groups:
            supplemental_group_ids = [
                (group.gr_gid) for group in grp.getgrall() if username in group.gr_mem
            ]

        return (passwd_entry, supplemental_group_ids)

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return user

    def check_user_allowed(self, username):
        """
        Check whether a user is allowed to create a collection, based on a
        storage_gateway's policies.
        """
        if username == "gcsweb":
            return False

        passwd_entry, groups = self.lookup_unix_info(username)
        if passwd_entry is not None and _utils is not None:
            group_id = passwd_entry.pw_gid
            private_policies = {}
            try:
                if self.users_allow:
                    private_policies["users_allow"] = ",".join(self.users_allow)
            except AttributeError:
                pass
            try:
                if self.users_deny:
                    private_policies["users_deny"] = ",".join(self.users_deny)
            except AttributeError:
                pass
            try:
                if self.groups_allow:
                    private_policies["groups_allow"] = ",".join(self.groups_allow)
            except AttributeError:
                pass
            try:
                if self.groups_deny:
                    private_policies["groups_deny"] = ",".join(self.groups_deny)
            except AttributeError:
                pass

            policy = dict(private=private_policies)
            current_app.logger.debug(
                "check_user_allowed: user=%s policy=%s", username, policy
            )
            res = _utils.globus_gfs_check_user_allowed_json(
                passwd_entry.pw_name.encode("utf8"),
                group_id,
                (ctypes.c_int * len(groups))(*groups),
                len(groups),
                json.dumps(policy).encode("utf8"),
            )

            if res:
                return True

        return False
