from urllib.parse import urlparse
from collections.abc import Sequence
from typing import List

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway
from globus.manager.constants import ConfigurationObjectTag


class S3StorageGateway(StorageGateway):
    CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = [
        "s3_endpoint",
        "s3_buckets",
        "s3_user_credential_required",
    ]

    connector = Connector(id=CONNECTOR_ID, display_name="S3")

    @property
    def s3_endpoint(self) -> str:
        """
        URL of the S3 API endpoint
        """
        value = self._s3_endpoint
        if value is None:
            raise AttributeError("s3_endpoint")

        return value

    @s3_endpoint.setter
    def s3_endpoint(self, value):
        self._s3_endpoint = value

    @checks_policy_property(prop="s3_endpoint", required=True)
    def check_s3_endpoint(self, value: str):
        if value is not None:
            try:
                parsed_value = urlparse(value)
                if parsed_value.scheme not in ["http", "https"]:
                    raise ValueError()
            except Exception:
                raise ValueError()
        else:
            raise ValueError()

    @property
    def s3_buckets(self) -> List[str]:
        """
        Optional list of S3 buckets which this storage gateway allows access
        to. If this is not set, then this Storage Gateway allows access to any
        buckets that User Credentials have access to.
        """
        value = self._s3_buckets
        if value is None:
            raise AttributeError("s3_buckets")

        return value

    @s3_buckets.setter
    def s3_buckets(self, value):
        self._s3_buckets = value

    @s3_buckets.deleter
    def s3_buckets(self):
        self._s3_buckets = None

    @checks_policy_property(prop="s3_buckets")
    def check_s3_buckets(self, value: List[str]):
        if value is not None:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    @property
    def s3_user_credential_required(self) -> bool:
        value = self._s3_user_credential_required
        if value is None:
            raise AttributeError("s3_user_credential_required")
        return value

    @s3_user_credential_required.setter
    def s3_user_credential_required(self, value):
        self._s3_user_credential_required = value
        self._s3_user_credential_available = value

    @checks_policy_property(prop="s3_user_credential_required", required=True)
    def check_s3_user_credential_required(self, value: bool):
        if value is None or not isinstance(value, bool):
            raise ValueError()

    @property
    def load_dsi_module(self):
        return "s3"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        pass

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass

    @property
    def require_credential(self):
        return self.s3_user_credential_required

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "s3_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "s3_storage_policies#1.0.0":
            raise ValueError("Unknown DATA_TYPE")

        super().patch_policies(**policies)

    def __repr__(self):
        return "<S3StorageGateway {}>".format(self.id)
