from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


GOOGLE_DRIVE_CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"


class GoogleDriveStoragePoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_drive_storage_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    client_id = ma.fields.Str(required=True, allow_none=True)
    secret = ma.fields.Str(required=True, allow_none=True)
    user_api_rate_quota = ma.fields.Int(required=False)
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        doc="URL of the auth callback that must be registered on the Google "
        "API console for the application client_id in order to process "
        "Google credentials.",
    )

    private = ["client_id", "secret", "user_api_rate_quota"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


class GoogleDriveCollectionPoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_drive_collection_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


class GoogleDriveUserCredentialPoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_drive_user_credential_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    sub = ma.fields.Str(allow_none=True)
    email = ma.fields.Email(allow_none=True)
    access_token = ma.fields.Str(required=False, allow_none=True)
    refresh_token = ma.fields.Str(required=False, allow_none=True)
    token_expiry = ma.fields.AwareDateTime()
    scopes = ma.fields.List(ma.fields.Str)

    private = ["access_token", "refresh_token"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


google_drive_user_credential_policies_private_schema = (
    GoogleDriveUserCredentialPoliciesSchema()
)
google_drive_user_credential_policies_schema = GoogleDriveUserCredentialPoliciesSchema(
    load_only=GoogleDriveUserCredentialPoliciesSchema.private
)
google_drive_storage_policies_private_schema = GoogleDriveStoragePoliciesSchema()
google_drive_storage_policies_schema = GoogleDriveStoragePoliciesSchema(
    load_only=GoogleDriveStoragePoliciesSchema.private
)
google_drive_collection_policies_schema = GoogleDriveCollectionPoliciesSchema()

DataTypeSchema.register_schema(google_drive_storage_policies_schema)
DataTypeSchema.register_schema(google_drive_user_credential_policies_schema)
DataTypeSchema.register_schema(google_drive_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=google_drive_storage_policies_schema,
    private=google_drive_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=google_drive_user_credential_policies_schema,
    private=google_drive_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(google_drive_collection_policies_schema)
