from flask import current_app, g
import globus_sdk


def log_auth_exception(exc: globus_sdk.exc.GlobusAPIError) -> None:
    """
    Log an exception from the Globus Auth Service
    """
    request_id = getattr(g, "request_id", None) if g else None
    raw_json = exc.raw_json or {}
    errors = raw_json.get("errors", [])
    auth_id = None
    for i in errors:
        if "id" in i:
            auth_id = i["id"]
            break

    current_app.logger.info(
        "auth_exception: gcs_request_id=%s "
        'auth_request_id=%s error=%s error_description="%s"',
        request_id,
        auth_id,
        raw_json.get("error"),
        raw_json.get("error_description"),
    )


def log_dns_exception(exc: globus_sdk.exc.GlobusAPIError) -> None:
    """
    Log an exception from the Globus DNS Service
    """
    request_id = getattr(g, "request_id", None) if g else None
    response = getattr(exc, "_underlying_response", None)
    headers = response.headers if response else {}
    dns_request_id = headers.get("globus_dns_request_id", None)
    raw_json = exc.raw_json or {}

    current_app.logger.info(
        "dns_exception: gcs_request_id=%s "
        'dns_request_id=%s error_code=%s message="%s"',
        request_id,
        dns_request_id,
        raw_json.get("error_code"),
        raw_json.get("message"),
    )


def log_groups_exception(exc: globus_sdk.exc.GlobusAPIError) -> None:
    """
    Log an exception from the Globus Groups Service
    """
    request_id = getattr(g, "request_id", None) if g else None
    response = getattr(exc, "_underlying_response", None)
    headers = response.headers if response else {}
    groups_request_id = headers.get("X-GO-Request-Id", None)
    raw_json = exc.raw_json or {}

    current_app.logger.info(
        "groups_exception: gcs_request_id=%s "
        'groups_request_id=%s code=%s message="%s"',
        request_id,
        groups_request_id,
        raw_json.get("code"),
        raw_json.get("message"),
    )


def log_transfer_exception(exc: globus_sdk.exc.TransferAPIError) -> None:
    """
    Log an exception from the Globus Transfer Service
    """
    request_id = getattr(g, "request_id", None) if g else None
    raw_json = exc.raw_json or {}
    current_app.logger.info(
        "transfer_exception: gcs_request_id=%s "
        'transfer_request_id=%s code=%s message="%s" resource=%s',
        request_id,
        getattr(exc, "request_id", None),
        getattr(exc, "code", None),
        getattr(exc, "message", None),
        raw_json.get("resource"),
    )
