import globus.manager.api.models as models


class SubscriptionIdChecker:
    """
    This class checks the Subscription ID of the non_functional host endpoint
    twice a day. This is not something that changes often, but it is possible
    that a user changes the subscription_id in transfer without using the GCS
    Manager API.
    """

    CHECK_DLEAY = 60 * 60 * 12
    """
    Delay (in seconds) to wait between endpoint subscription id check.
    """

    def __init__(self, scheduler, app):
        """
        Initialize a Subscription object to handle the subscription id
        synchronization.

        :param sched.scheduler scheduler:
            A scheduler which this class will use to register timed events.
        :param GCSManagerApp app:
            An app which this class will use to interact with appsync, the
            database, locks and logs.
        """
        self.scheduler = scheduler
        "Scheduler to register events with"

        self.app = app
        "Flask application this object is a part of"

        self.appsync_client = app.appsync_client

        self.lock = self.appsync_client.subscription_lock
        "Subscription lock to synchronize with other appsync access"

        self.scheduler.enter(60, 0, self.check_subscription_id)

    def check_subscription_id(self):
        """
        Event to check the subscription_id property of the non-functional
        host endpoint in transfer. If the subscription ID does not match the
        value currently in the database, update the database value and push
        the change to appsync.

        This is triggered when the Subscription is first started, and
        then twice a day.

        If there's an error checking the endpoint, it is ignored and the next
        check is scheduled with the normal check interval.
        """
        with self.app.app_context(), self.lock.exclusive():
            self.app.logger.debug("check_subscription_id: enter")
            old_subscription = None
            new_subscription_id = None
            endpoint_id = None
            try:
                endpoint = models.Endpoint.query.one_or_none()
                old_subscription = getattr(endpoint, "subscription_id", None)
                endpoint_id = endpoint.endpoint_id
                self.app.logger.debug(
                    "check_subscription_id: endpoint=%s, subscription_id=%s",
                    endpoint_id,
                    old_subscription,
                )
            except Exception as e:
                self.app.logger.warning(
                    "check_subscription_id: Error checking endpoint subscription_id: %s",
                    str(e),
                )
                pass
            if endpoint_id:
                try:
                    res = self.app.transfer_client.get_endpoint(endpoint_id).data
                    new_subscription_id = res.get("subscription_id")
                    self.app.logger.debug(
                        "check_subscription_id: new_subscription_id = %s",
                        new_subscription_id,
                    )
                    if old_subscription != new_subscription_id:
                        self.app.logger.debug(
                            "check_subscription_id: updating appsync endpoint"
                        )
                        endpoint.subscription_id = new_subscription_id
                        self.appsync_client.update_object(endpoint)
                except Exception as e:
                    self.app.logger.warn(
                        "check_subscription_id: Unable to check subscription for %s: %s",
                        endpoint_id,
                        str(e),
                    )
                    pass
            self.app.logger.debug("check_subscription_id: exit")

        self.scheduler.enter(self.CHECK_DLEAY, 0, self.check_subscription_id)
