import json
import os
import textwrap

import click
from flask import current_app
from jwcrypto import jwk

import globus_sdk
from gcs_cli.parsing.command_decorator import globus_command

from globus.manager.version import __version__
from globus.manager.api.app import make_app
import globus.manager.api.models as models
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import ConfigurationObjectTag


class NodeKeyFileParam(click.File):
    name = "node_key_file"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        with super().convert(value, param, ctx) as f:
            try:
                node_key = jwk.JWK(**json.loads(f.read()))
            except Exception as e:
                self.fail("Error parsing {}: {}".format(value, str(e)), param, ctx)
        return node_key


@globus_command(
    "cleanup",
    short_help="Clean up a Globus Connect Server endpoint",
    help="Clean up a Globus Connect Server endpoint",
    disable_options=["format", "explicit_host"],
)
@click.help_option("--help", "-h")
@click.version_option(version=__version__, prog_name="globus-connect-server")
@click.option(
    "--client-id",
    "-c",
    help="Auth Client ID for this Globus Connect Server Endpoint",
    required=True,
)
@click.option(  # type: ignore # hidden/required
    "--secret",
    "-s",
    help="Secret for the auth client",
    required=True,
    hidden=True,
    prompt=True,
    hide_input=True,
)
@click.option(
    "--deployment-key", help="Path to deployment key", type=NodeKeyFileParam(),
)
@click.pass_context
def endpoint_cleanup(ctx, client_id, secret, deployment_key):
    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    models.db.init_app(app)
    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "default")

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()
        if deployment_key:
            print("Loaded deployment key", deployment_key.thumbprint())
            app.config["DEPLOYMENT_KEY"] = deployment_key
        current_app.config["CLIENT_ID"] = client_id
        current_app.config["CLIENT_SECRET"] = secret
        current_app.config["ENVIRONMENT"] = environment

        if not deployment_key:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "You did not provide the --deployment-key command-line "
                        "option when invoking this command. Not all resources "
                        "can not be deleted without a deployment key. If you "
                        "still have access to that key, please rerun this command "
                        "with that option."
                    )
                )
            )
            if click.confirm("Do you still access to your deployment key?"):
                ctx.exit(0)

        click.echo(
            "\n".join(
                textwrap.wrap(
                    "This command will completely remove all configuration "
                    "information associated with this endpoint. This is "
                    "irreversible, and the client_id used to create the "
                    "endpoint may not be reused to recreate this endpoint.",
                    width=78,
                )
            )
        )

        if not click.confirm("Do you wish to continue?"):
            ctx.exit(1)

        command = EndpointDelete(ctx, client_id, secret, deployment_key, environment)
        command.run()


class EndpointDelete:
    def __init__(self, ctx, client_id, secret, deployment_key, environment):
        self.ctx = ctx
        self.client_id = client_id
        self.secret = secret
        self.deployment_key = deployment_key
        self.environment = environment
        self.endpoint_id = client_id
        self.items = None

    def synchronize_state(self):
        "Synchronize endpoint state"
        kc = None

        if self.deployment_key:
            kc = current_app.appsync_client.query_all(
                id=ConfigurationObjectTag.KEY_CHAIN.value
            )
            for k in kc:
                current_app.appsync_client.process_item(k)

        try:
            self.items = current_app.appsync_client.query_all()
        except GraphQLError:
            self.items = []

        for it in self.items:
            if it["tag"] == ConfigurationObjectTag.ENDPOINT:
                if self.deployment_key:
                    self.endpoint_id = current_app.keychain.decrypt(
                        it["configuration"]
                    )["id"]

        if kc:
            current_app.appsync_client.commit_items_to_session(
                models.db.session, self.items
            )

    def check_nodes(self):
        "Check for existing node"
        domain = models.Domain.query.one_or_none()

        if domain:
            existing_nodes = []
            for node in domain.nodes:
                existing_nodes.append(node)

            if existing_nodes:
                text = "\n".join(
                    ["\n"]
                    + textwrap.wrap(
                        "There are still node(s) configured to serve "
                        "this endpoint running on the following IP addresses:",
                        width=78,
                    )
                    + [""]
                    + [
                        "    " + ", ".join([str(n) for n in node.ip_addresses])
                        for node in existing_nodes
                    ]
                    + [""]
                    + textwrap.wrap(
                        "If you still have access to these nodes, log into "
                        "them and run the following command as root to remove "
                        "them from the endpoint and disable the Globus services "
                        "on them:",
                        width=78,
                    )
                    + ["\n    globus-connect-server node cleanup"]
                    + [""]
                )
                click.echo(text)
                if not click.confirm("Continue without cleaning up nodes?"):
                    self.ctx.exit(1)

    def delete_collections(self):
        "Delete collections"
        for collection in models.Collection.query.filter(
            ~models.Collection.deleted, models.Collection.is_guest_collection
        ):
            try:
                collection.delete()
            except globus_sdk.GlobusAPIError as e:
                if e.http_status != 404:
                    current_app.logger.debug("Error deleting collection %s", str(e))

        for collection in models.Collection.query.filter(
            ~models.Collection.deleted, models.Collection.is_mapped_collection
        ):
            try:
                collection.delete()
            except globus_sdk.GlobusAPIError as e:
                if e.http_status != 404:
                    current_app.logger.warn("Unable to delete collection %s", str(e))

        for ep in current_app.transfer_client.endpoint_search(
            filter_scope="my-endpoints", num_results=None
        ):
            if not ep.get("non_functional"):
                try:
                    current_app.transfer_client.delete_endpoint(ep["id"])
                except globus_sdk.GlobusAPIError as e:
                    if e.http_status != 404:
                        current_app.logger.warn(
                            "Unable to delete collection %s", str(e)
                        )
            elif not self.endpoint_id:
                self.endpoint_id = ep["id"]

    def delete_user_credentials(self):
        "Delete user credentials"
        for uc in models.UserCredential.query.filter(~models.UserCredential.deleted):
            try:
                uc.delete()
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete user credential %s: %s", uc, str(e)
                )

    def delete_domain(self):
        "Delete DNS name"
        domain = models.Domain.query.one_or_none()
        if domain:
            try:
                domain.delete()
            except Exception as e:
                current_app.logger.warn("Unable to delete DNS name: %s", str(e))

    def delete_endpoint(self):
        "Delete endpoint"
        try:
            if self.endpoint_id:
                current_app.transfer_client.delete_endpoint(self.endpoint_id)
        except globus_sdk.TransferAPIError as e:
            if e.http_status != 404:
                current_app.logger.warn("Unable to delete endpoint: %s", str(e))

    def delete_auth_clients(self):
        "Delete virtual clients"
        clients = []
        try:
            result = current_app.auth_client.get("/v2/api/clients")
            clients = result.data.get("clients")
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.warning(e)

        for client in clients:
            try:
                client_id = client["id"]
                if client["id"] != self.client_id:
                    current_app.auth_client.delete(
                        "/v2/api/clients/{}".format(client_id)
                    )
            except Exception as e:
                current_app.logger.warning(e)

    def delete_appsync_data(self):
        "Delete appsync data"

        endpoint_object = None
        keychain_object = None

        all_deleted = True

        for item in self.items:
            if item["tag"] == ConfigurationObjectTag.ENDPOINT:
                endpoint_object = item
            elif item["tag"] == ConfigurationObjectTag.KEY_CHAIN:
                keychain_object = item
            else:
                try:
                    current_app.appsync_client.delete_object(id=item["id"])
                except Exception as e:
                    current_app.logger.warn(
                        "Unable to delete %s object %s: %s",
                        item["tag"].value,
                        item["id"],
                        str(e),
                    )
                    all_deleted = False

        if all_deleted and keychain_object:
            try:
                current_app.appsync_client.delete_object(id=keychain_object["id"])
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete %s object %s: %s",
                    keychain_object["tag"].value,
                    keychain_object["id"],
                    str(e),
                )
                all_deleted = False

        if all_deleted and endpoint_object:
            try:
                current_app.appsync_client.delete_object(id=endpoint_object["id"])
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete %s object %s: %s",
                    endpoint_object["tag"].value,
                    endpoint_object["id"],
                    str(e),
                )
                all_deleted = False
        if not all_deleted:
            current_app.logger.info("Not all objects deleted cleanly")

    def run(self):
        tasks = [
            self.synchronize_state,
            self.check_nodes,
            self.delete_collections,
            self.delete_user_credentials,
            self.delete_domain,
            self.delete_endpoint,
            self.delete_auth_clients,
            self.delete_appsync_data,
        ]

        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show) as bar:
            for task in bar:
                task()
