"""
globus-connect-server endpoint migrate53
"""
import os
import pwd
import textwrap

import click
from flask import current_app
from gcs_cli.parsing.command_decorator import globus_command
import globus_sdk

from .configuration_data import ConfigurationDataV53

from globus_gcs.utils.host_info import public_ip

from globus_gcs.apache.server import WebServerManager
from globus.manager.api.blueprints.collections import allocate_domain
from globus.manager.api.blueprints.utils import create_object
import globus.manager.api.models as models
from globus.manager.api.schemas.api import endpoint_schema
from globus.manager.api.schemas.api.collection import collection_1_1_0_private_schema
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import CollectionType, ConfigurationObjectTag
from globus.manager.scripts.endpoint_setup import EndpointSetup
from globus.manager.scripts.node_setup import NodeSetup
from globus.manager.version import __version__

from globus.manager.api.app import make_app


@globus_command(
    "migrate53",
    short_help="Migrate a Globus Connect Server v5.3",
    help="Migrate a Globus Connect Server v5.3 endpoint for use with v5.4",
    disable_options=["format", "explicit_host"],
)
@click.help_option("--help", "-h")
@click.version_option(version=__version__, prog_name="globus-connect-server")
@click.option(
    "--deployment-key",
    "-d",
    help="Path for deployment key configuration.",
    type=click.Path(exists=False),
    default=os.path.expanduser("~gcsweb/deployment-key.json"),
)
@click.option(
    "--owner",
    "-o",
    help="Globus username of the owner of this endpoint. This user will be "
    "granted the owner role for this endpoint.",
    required=True,
)
@click.option("--ip-address", "-i", help="IP Address of this data transfer node")
@click.option(
    "--root",
    hidden=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="/",
)
def migrate53_command(deployment_key, owner, ip_address, root):
    ctx = click.get_current_context()

    try:
        gcsweb_pwd = pwd.getpwnam("gcsweb")
    except KeyError:
        click.echo(
            "This node does not have the gcsweb user created... aborting", err=True
        )
        ctx.exit(1)

    if os.getuid() != 0:
        click.echo("This script must be run as root", err=True)
        ctx.exit(1)

    if ip_address is None:
        ip_address = public_ip()
        click.echo("IP Address not specified, using {}".format(ip_address))

    WebServerManager.site_remove("gcs_manager")

    # Create endpoint app before setuid, so that the log is owned by root
    endpoint_app = make_app(app_config="globus.manager.api.config.endpoint_cli")

    # We use the endpoint app for the migration steps where we create the
    # appsync objects from the v5.3 endpoint configuration
    models.db.init_app(endpoint_app)
    with endpoint_app.app_context():
        models.db.create_all()
        models.db.session.commit()

        endpoint_migrate = EndpointMigrate(ctx, endpoint_app, deployment_key, owner)
        endpoint_setup = EndpointSetup(
            ctx,
            endpoint_app.config["CLIENT_ID"],
            endpoint_app.config["CLIENT_SECRET"],
            deployment_key,
            owner,
            endpoint_schema.dump(endpoint_migrate.configuration_data.endpoint),
            endpoint_app.config["ENVIRONMENT"],
        )

        click.echo("Preparing")
        os.setegid(gcsweb_pwd.pw_gid)
        os.seteuid(gcsweb_pwd.pw_uid)

        try:
            endpoint_setup.load_or_create_deployment_key()
        except PermissionError:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "The deployment key path {} must be readable and "
                        "writable by the gcsweb user".format(deployment_key)
                    )
                ),
                err=True,
            )
            ctx.exit(1)

        endpoint_migrate.ensure_stub_endpoint()
        endpoint_setup.check_for_existing_keychain()
        endpoint_setup.add_keys_to_keychain()
        endpoint_migrate.run_pre()

        click.echo("Setting up endpoint")
        endpoint_setup.run()

        click.echo("Migrating configuration")
        endpoint_migrate.run_post()

        if endpoint_app.config["ENVIRONMENT"] == "default":
            webapp = "https://app.globus.org"
        elif endpoint_app.config["ENVIRONMENT"] == "preview":
            webapp = "https://app.preview.globus.org"
        else:
            webapp = "https://app.{}.globuscs.info".format(
                endpoint_app.config["ENVIRONMENT"]
            )

        migration_artifact_collections = {}

        for existing in models.Collection.query.filter(
            ~models.Collection.deleted, models.Collection.is_mapped_collection
        ):
            keywords = getattr(existing, "keywords", [])
            for k in keywords:
                if k.startswith("migration:"):
                    migration_artifact_collections[
                        existing.id
                    ] = "{}/file-manager/collections/{}/overview/edit".format(
                        webapp, existing.id
                    )
                    break

        if migration_artifact_collections:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "As part of this migration, this script created "
                        "{} mapped collection(s) to support changes in "
                        "the collection model for v5.4. To help your users "
                        "find the collections on your endpoint, please visit "
                        "the following URLs to update the metadata and names "
                        "of those collections".format(
                            len(migration_artifact_collections.keys())
                        )
                    )
                    + [v for v in migration_artifact_collections.values()]
                )
            )

    # Now we use a node app for the migration steps where we configure the
    # current node to serve this endpoint
    node_app = make_app(app_config="globus.manager.api.config.node_cli")
    models.db.init_app(node_app)
    with node_app.app_context():
        models.db.create_all()
        models.db.session.commit()

        node_app.config["DEPLOYMENT_KEY"] = endpoint_app.config["DEPLOYMENT_KEY"]
        node_app.config["CLIENT_ID"] = endpoint_app.config["CLIENT_ID"]
        node_app.config["CLIENT_SECRET"] = endpoint_app.config["CLIENT_SECRET"]
        node_app.config["ENVIRONMENT"] = endpoint_app.config["ENVIRONMENT"]
        node_app.config["HOST_ENDPOINT_ID"] = endpoint_app.config["HOST_ENDPOINT_ID"]
        node_app.keychain.add_node_key(endpoint_app.config["DEPLOYMENT_KEY"])

        setup_node = NodeSetup(
            ctx,
            node_app.config["CLIENT_ID"],
            node_app.config["CLIENT_SECRET"],
            node_app.config["DEPLOYMENT_KEY"],
            endpoint_migrate.configuration_data.incoming_port_range,
            endpoint_migrate.configuration_data.outgoing_port_range,
            ip_address,
            node_app.config["ENVIRONMENT"],
            None,
            None,
            root,
        )

        click.echo("Configuring services")
        setup_node.run_non_root()
        os.seteuid(0)
        os.setegid(0)

        click.echo("Cleaning up old configuration")
        endpoint_migrate.run_post_root()

        click.echo("Starting services")
        setup_node.run_root()

        click.echo(
            "\n".join(
                textwrap.wrap(
                    "Migrated endpoint {}.\nYou'll need the following information "
                    "to deploy additional nodes, as well as the contents of {}:".format(
                        node_app.config["HOST_ENDPOINT_ID"],
                        deployment_key if deployment_key else "deployment-key.json",
                    )
                )
                + [
                    "\tclient_id: {}".format(node_app.config["CLIENT_ID"]),
                    "\tsecret: {}".format(node_app.config["CLIENT_SECRET"]),
                ]
            )
        )


class EndpointMigrate:
    def __init__(self, ctx, app, deployment_key, owner):
        self.configuration_data = ConfigurationDataV53(owner)
        configuration = self.configuration_data.configuration
        self.owner = owner

        # Load configuration table into the v5.4 database and add its values
        # to the app config
        for name, val in configuration.items():
            if name == "environment":
                val = val.lower()
            app.config[name.upper()] = val
            cfg = models.Configuration.query.filter(
                models.Configuration.name == name
            ).one_or_none()
            if cfg:
                cfg.value = val
            else:
                cfg = models.Configuration(name, val)
            models.db.session.add(cfg)
        models.db.session.commit()

        # Temporary value used when getting the new cert
        app.config["NODE_ID"] = app.config["CLIENT_ID"]

        try:
            self.appsync_client = app.appsync_client
            if self.appsync_client is None:
                click.echo("Unable to contact Globus services", err=True)
                ctx.exit(1)
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.error(e)
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "Unable to contact Globus services with provided client "
                        "id and secret. Please verify that those values are "
                        "correct."
                    )
                ),
                err=True,
            )
            ctx.exit(1)

    def _migrate_object(self, obj):
        try:
            current_app.appsync_client.create_object(obj)
        except GraphQLError as e:
            current_app.logger.error(
                "%s %s migration error %s", obj.tag.value, obj.id, e.message
            )

    def migrate_owner_id(self):
        """
        Verify owner ID
        """
        current_app.logger.info("verify_owner_id: start")
        owner_identities = current_app.auth_client.get_identities(
            usernames=[self.owner]
        ).data.get("identities", [])
        if len(owner_identities) != 1:
            click.echo(
                "\nUnable to verify that Globus user {} exists".format(self.owner)
            )
            self.ctx.exit(1)

        owner_identity = owner_identities[0]
        current_app.logger.debug(
            "verify_owner_id: username=%s id=%s status=%s",
            self.owner,
            owner_identity["id"],
            owner_identity["status"],
        )
        if owner_identity["status"] != "used":
            click.echo(
                "\nRequested owner {} exists but does not have a "
                "validated account".format(self.owner)
            )
            self.ctx.exit(1)
        self.owner_id = owner_identity["id"]
        current_app.logger.info("verify_owner_id: end")

    def migrate_dns(self):
        """
        Replace DNS domain with
        """
        old_domain = self.configuration_data.domain
        if old_domain:
            exists_in_appsync = False
            try:
                for item in current_app.appsync_client.query_all(
                    tag=ConfigurationObjectTag.WILDCARD_DOMAIN
                ):
                    exists_in_appsync = True
                    item_config = current_app.keychain.decrypt(item["configuration"])
                    if item_config["domain_name"] == old_domain.domain_name:
                        old_domain.update(item)
                    else:
                        msg = "Migration error: endpoint domain change {} to {}".format(
                            old_domain.domain_name, item_config["domain_name"]
                        )

                        current_app.logger.error(msg)
                        raise Exception(msg)
            except GraphQLError as e:
                if e.errors[0].get("errorType") == "CONFLICT":
                    pass
                else:
                    raise

            old_domain.update_domain()
            if not exists_in_appsync:
                current_app.appsync_client.create_object(old_domain)

    def migrate_core_objects(self):
        # Migrate the endpoint, acme_account, and endpoint_roles into appsync
        try:
            self.configuration_data.endpoint.refresh_configuration_hash()
            current_app.appsync_client.update_object(self.configuration_data.endpoint)
            models.db.session.add(self.configuration_data.endpoint)
            models.db.session.commit()
        except GraphQLError as e:
            if e.data:
                res = e.data
                for o in res.get("configuration_objects", {}).get("items", []):
                    if (
                        ConfigurationObjectTag(o["tag"])
                        == ConfigurationObjectTag.ENDPOINT
                        and o["id"] == current_app.config["HOST_ENDPOINT_ID"]
                    ):
                        self.configuration_data.endpoint.update(o)
                        models.db.session.add(self.configuration_data.endpoint)
                        models.db.session.commit()
                        break
        self.configuration_data.acme_account.refresh_configuration_hash()
        self._migrate_object(self.configuration_data.acme_account)

        for role in self.configuration_data.role.values():
            role.refresh_configuration_hash()
            if role.resource is None:
                self._migrate_object(role)

    def migrate_storage_gateways(self):
        "Update storage gateways"
        for sg in self.configuration_data.storage_gateway.values():
            sg.refresh_configuration_hash()
            self._migrate_object(sg)

    def migrate_user_credentials(self):
        "Update user credentials"
        for uc in self.configuration_data.user_credential.values():
            uc.refresh_configuration_hash()
            self._migrate_object(uc)

    def migrate_collections(self):
        """
        Migrate collections
        """
        # These collections are being created as part of the migration process
        synthesized_mapped_collections = [
            co
            for co in self.configuration_data.collection.values()
            if co.id.startswith(self.configuration_data.UNDETERMINED_UUID_PREFIX)
        ]

        items = current_app.appsync_client.query_all()
        deferred = current_app.appsync_client.commit_items_to_session(
            models.db.session, items
        )

        for mc in synthesized_mapped_collections:
            # Create mapped collection and roles (updates auth, transfer, appsync)
            # if they don't exist yet. We stash `migration:$storage_gateway_id`
            # in keywords so if this is run twice it will not recreate things.
            migration_keyword = "migration:{}".format(mc.storage_gateway_id)
            collection = None
            for existing in models.Collection.query.filter(
                ~models.Collection.deleted,
                models.Collection.is_mapped_collection,
                models.Collection.storage_gateway_id == mc.storage_gateway_id,
            ):
                if migration_keyword in getattr(existing, "keywords", []):
                    collection = existing
                    break

            if not collection:
                mc_doc = collection_1_1_0_private_schema.dump(mc)
                collection_type = CollectionType(mc_doc["collection_type"])
                mc_doc = {
                    k: v
                    for k, v in mc_doc.items()
                    if k in collection_1_1_0_private_schema.fields
                    and not collection_1_1_0_private_schema.fields[k].dump_only
                }
                mc_doc["collection_type"] = collection_type
                sg = models.StorageGateway.query.filter(
                    models.StorageGateway.id == mc.storage_gateway_id
                ).one()

                domain_name = allocate_domain(mc.collection_type)
                endpoint = models.Endpoint.query.one()
                try:
                    collection = models.Collection.from_collections_post(
                        storage_gateway=sg,
                        endpoint=endpoint,
                        domain_name=domain_name,
                        authorized_identity={"id": self.owner_id},
                        doc=mc_doc,
                    )
                except Exception as e:
                    click.echo(
                        "\n".join(
                            textwrap.wrap(
                                "Error creating collection {}. Migration will continue, "
                                "but some migrated collections will be unavailable. Please "
                                "contact globus support for help fixing this collection. ".format(
                                    mc_doc["display_name"]
                                ),
                            )
                            + ["Exception: " + str(e)]
                        ),
                        err=True,
                    )
                    continue
                create_object(collection)
                if collection.owner_role:
                    create_object(collection.owner_role)

            # Update our migration model to match the actual transfer uuid
            old_id = mc.id
            self.configuration_data.collection.pop(old_id)
            mc.id = collection.id
            self.configuration_data.collection[mc.id] = mc

            # Update any guest collections that were pointing to the fake uuid
            for gc in self.configuration_data.collection.values():
                if (
                    gc.collection_type == CollectionType.guest
                    and gc.mapped_collection_id == old_id
                ):
                    gc.mapped_collection_id = mc.id

        # Create models for the collections that already exist in transfer,
        # if the models aren't in appsync yet.
        for co in self.configuration_data.collection.values():
            if co in synthesized_mapped_collections:
                continue

            existing = models.Collection.query.filter(
                models.Collection.id == co.id
            ).one_or_none()

            # Previous version would omit mapped_collection_id if multiple
            # guests were created on a synthesized mapped collection, but
            # that would get uploaded into appsync. If that exists, the appsync
            # data will exist but be unparseable
            deferred_items_with_id = [i for i in deferred if i["id"] == co.id]
            if not existing and deferred_items_with_id:
                item = deferred_items_with_id[0]
                item_config = current_app.keychain.decrypt(item["configuration"])
                item_config["mapped_collection_id"] = co.mapped_collection_id
                item["configuration"] = current_app.keychain.encrypt(token=item_config)
                check = current_app.appsync_client.commit_items_to_session(
                    models.db.session, [item]
                )
                assert not check
                existing = models.Collection.query.filter(
                    models.Collection.id == co.id
                ).one()
                current_app.appsync_client.update_object(existing)
            elif not existing:
                if co.collection_type == CollectionType.mapped:
                    co.identity_id = self.owner_id
                co.domain_name = allocate_domain(co.collection_type)
                self._migrate_object(co)
                # Create the model from the appsync response (stashed in
                # col.appsync_data
                cls = models.Collection.polymorphic_class(co.connector_id)
                if co.mapped_collection_id:
                    mapped_collection = models.Collection.query.filter(
                        models.Collection.id == co.mapped_collection_id
                    ).one()
                else:
                    mapped_collection = None
                endpoint = models.Endpoint.query.one()
                try:
                    existing = cls(
                        appsync_data=co.appsync_data,
                        storage_gateway=co.storage_gateway,
                        mapped_collection=mapped_collection,
                        user_credential=co.user_credential,
                        endpoint=endpoint,
                    )
                except Exception as e:
                    click.echo(
                        "\n".join(
                            textwrap.wrap(
                                "Error migrating collection {}. Migration will continue, "
                                "this migrated collections will be unavailable. Please "
                                "contact globus support for help fixing this collection.".format(
                                    co.display_name
                                ),
                            )
                            + ["Exception: " + str(e)]
                        ),
                        err=True,
                    )
                    continue
                models.db.session.add(endpoint)
            # Update transfer enpdoint so https/tlsftp URLs are updated
            existing.update_transfer_endpoint()

    def migrate_roles(self):
        "Create new roles"
        # Create models for roles
        for ro in self.configuration_data.role.values():
            existing = models.Role.query.filter(models.Role.id == ro.id).one_or_none()
            if not existing:
                self._migrate_object(ro)

    def migrate_clients(self):
        """
        Update auth clients
        """
        for c in models.Collection.query.filter(~models.Collection.deleted):
            c.update_child_client()

    def migrate_scopes(self):
        """
        Update auth scopes
        """
        endpoint = models.Endpoint.query.one()
        endpoint._update_endpoint_in_globus_services(
            create=False, changed_display_name=True
        )
        for c in models.Collection.query.filter(~models.Collection.deleted):
            c.update_scopes()

    def migrate_fingerprint(self):
        """
        Update endpoint version
        """
        endpoint = models.Endpoint.query.one()
        current_app.transfer_client.update_endpoint(
            endpoint.endpoint_id,
            {
                "gcs_private_version_json": (
                    self.configuration_data.endpoint.fingerprint
                )
            },
        )

    def __run(self, tasks):
        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.strip().split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show, width=25) as bar:
            for task in bar:
                task()

    def migrate_remove_unused_state(self):
        """
        Remove unused state
        """

        # vhosts may exist if they refer to a preallocated collection or
        # one which was deleted in transfer but not GCS.
        stale_vhosts = []
        vhost_dir = os.path.expanduser("~gcsweb/etc/httpd/conf.d")

        for vhost in os.listdir(vhost_dir):
            vhost_id = vhost.split(".conf", 1)[0]
            collection = models.Collection.query.filter(
                models.Collection.id == vhost_id,
                ~models.Collection.deleted,  # type: ignore
            ).one_or_none()
            if not collection:
                stale_vhosts.append(vhost)

        for stale in stale_vhosts:
            stale_path = os.path.join(vhost_dir, stale)
            if os.path.exists(stale_path):
                os.remove(stale_path)

        # GCSv5.3 adds these files to avoid duplicating (the DB interface
        # probably should have been used for that instead). They are not
        # used for anything.
        gcs_dir = os.path.expanduser("/var/lib/globus-connect-server")
        for unused in ["dns-name.txt", "endpoint-uuid.txt"]:
            unused_path = os.path.join(gcs_dir, unused)
            if os.path.exists(unused_path):
                os.remove(unused_path)

    def ensure_stub_endpoint(self):
        """
        Ensure that at least a stub endpoint object exists in appsync. There
        needs to be at least a stub to be able to quired to be able to query
        anything related to the endpoint in appsync. Otherwise, they KeyChain
        becomes inaccessible.
        """
        try:
            items = current_app.appsync_client.query_all(
                tag=ConfigurationObjectTag.ENDPOINT
            )
            if items:
                return
        except GraphQLError as e:
            if e.errors[0].get("errorType") != "NOT_FOUND":
                raise

        stub_endpoint = models.Endpoint(doc={})
        current_app.appsync_client.create_object(stub_endpoint)

    def run_pre(self):
        tasks = [
            self.migrate_owner_id,
            self.migrate_core_objects,
            self.migrate_dns,
        ]
        self.__run(tasks)

    def run_post(self):
        tasks = [
            self.migrate_storage_gateways,
            self.migrate_user_credentials,
            self.migrate_collections,
            self.migrate_roles,
            self.migrate_clients,
            self.migrate_scopes,
            self.migrate_fingerprint,
        ]
        self.__run(tasks)

    def run_post_root(self):
        tasks = [
            self.migrate_remove_unused_state,
        ]
        self.__run(tasks)
