import re
from typing import Any, Dict, List, Optional

import globus.manager.api.models as models
from globus.manager.api.schemas.api import storage_gateway_private_schema
from globus.manager.utils.path_restriction import PathRestrictions


class _StorageGatewayV53:
    """
    Base class of a storage gateway exported from a GCSv5.3 installation. This
    maps the properties of the storage gateway into something which can be
    dumped using the v5.4 marshmallow schema. It also contains some private
    fields which do not directly map to v5.4 but will be needed to figure out
    what types of storage gateways will be needed to process the collections
    on this storage gateway.
    """

    model_class = models.StorageGateway

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self.id = kwargs["id"]
        self.display_name = kwargs["display_name"]
        self.connector_id = kwargs.get("connector_id")
        self.high_assurance = kwargs.get("require_high_assurance", False)
        self.deleted = kwargs.get("deleted", False)

        # These are stored, but are only used internally and are not visible
        # to the outside
        self._created_at = kwargs["created_at"]
        self._updated_at = kwargs["updated_at"]

        public_policies = kwargs.get("storage_policies", {}).get("public", {})
        private_policies = kwargs.get("storage_policies", {}).get("private", {})
        dsi_config = private_policies.get("dsi_config", {})
        credential = kwargs.get("credential", {}) or {}

        self._allow_guest_collections = kwargs.get("allow_guest_collections", False)
        self._allow_mapped_collections = kwargs.get("allow_mapped_collections", False)

        self.init_public_policies(**public_policies)
        self.init_private_policies(**private_policies)
        self.init_dsi_config(**dsi_config)
        self.init_credential(**credential)

    def _parse_path_restrictions(
        self, restrictions: Optional[str]
    ) -> Optional[PathRestrictions]:
        read = []
        read_write = []
        none = []
        if restrictions:
            for restriction in re.split(r"(?<!\\),", restrictions):
                m = re.match("^(R|RW|N)([/$].*)", restriction)
                if not m:
                    raise Exception("Bad Path Restriction")

                if m.group(1) == "R":
                    read.append(m.group(2))
                elif m.group(1) == "RW":
                    read_write.append(m.group(2))
                elif m.group(1) == "N":
                    none.append(m.group(2))
            return PathRestrictions(read=read, read_write=read_write, none=none)
        return None

    def init_public_policies(self, **policies):
        # This does not exist in v5.4, but can be used when synthesizing
        # mapped collections for storage gateways that have guest collections
        # but no mapped collections
        self._root = policies.get("root", "/")

        self.allowed_domains = policies.get("domains", "/")

        # XXX: identity_providers
        # Adding explicit list of identity providers is not supported in v5.4
        # Verify that the IdP list matches the allowed_domains value

        self.authentication_timeout_mins = policies.get(
            "authentication_assurance_timeout", 11 * 24 * 60
        )

    def init_private_policies(self, **policies):
        self.restrict_paths = self._parse_path_restrictions(
            policies.get("restrict_paths")
        )
        self.process_user = policies.get("process_user")
        self.load_dsi_module = policies.get("load_dsi_module")

    def init_dsi_config(self, **dsi_config):
        pass

    def init_credential(self, **credential):
        pass


class PosixStorageGatewayV53(_StorageGatewayV53):
    model_class = models.PosixStorageGateway

    def init_private_policies(
        self,
        users_allow: Optional[str] = None,
        users_deny: Optional[str] = None,
        groups_allow: Optional[str] = None,
        groups_deny: Optional[str] = None,
        **kwargs
    ):
        if users_allow:
            self.users_allow = users_allow.split(",")
        if users_deny:
            self.users_deny = users_deny.split(",")
        if groups_allow:
            self.groups_allow = groups_allow.split(",")
        if groups_deny:
            self.groups_deny = groups_deny.split(",")

        super().init_private_policies(**kwargs)


class GoogleCloudStorageGatewayV53(_StorageGatewayV53):
    model_class = models.GoogleCloudStorageGateway

    def init_dsi_config(self, **kwargs):
        self._init_dsi_config(**kwargs)

    def _init_dsi_config(
        self,
        google_cloud_storage_buckets: List[str] = None,
        google_cloud_storage_projects: List[str] = None,
        **kwargs
    ):
        if google_cloud_storage_buckets:
            self.buckets = google_cloud_storage_buckets
        if google_cloud_storage_projects:
            self.projects = google_cloud_storage_projects

        super().init_dsi_config(**kwargs)

    def init_credential(self, client_id: str = None, secret: str = None, **kwargs):
        self.client_id = client_id
        self.secret = secret
        super().init_credential(**kwargs)


class GoogleDriveStorageGatewayV53(_StorageGatewayV53):
    model_class = models.GoogleDriveStorageGateway

    def init_private_policies(self, **kwargs):
        self._init_private_policies(**kwargs)

    def _init_private_policies(self, load_dsi_module: str, **kwargs):
        dsi, *args = load_dsi_module.split(":", 1)
        if args:
            m = re.match("api_quota_rate=([1-9][0-9]*)", args[0])
            if m:
                self.user_api_rate_quota = m.group(1)
        super().init_private_policies(load_dsi_module=dsi, **kwargs)

    def init_credential(self, client_id: str = None, secret: str = None, **kwargs):
        self.client_id = client_id
        self.secret = secret
        super().init_credential(**kwargs)


class S3StorageGatewayV53(_StorageGatewayV53):
    model_class = models.S3StorageGateway

    def init_public_policies(self, s3_user_credential_required: bool = None, **kwargs):
        self.s3_user_credential_required = s3_user_credential_required
        super().init_public_policies(**kwargs)

    def init_dsi_config(self, **kwargs):
        self._init_dsi_config(**kwargs)

    def _init_dsi_config(
        self,
        s3_endpoint: str,
        s3_buckets: List[str] = None,
        s3_user_credential_available: bool = None,
        **kwargs
    ):
        assert s3_user_credential_available == self.s3_user_credential_required
        self.s3_endpoint = s3_endpoint
        if s3_buckets:
            self.s3_buckets = s3_buckets
        super().init_dsi_config(**kwargs)


class CephStorageGatewayV53(_StorageGatewayV53):
    model_class = models.CephStorageGateway

    def init_public_policies(self, **kwargs):
        super().init_public_policies(**kwargs)

    def init_dsi_config(
        self,
        s3_endpoint: str = None,
        s3_buckets: List[str] = None,
        ceph_admin_key_id: str = None,
        ceph_admin_secret_key: str = None,
        **kwargs
    ):
        self.s3_endpoint = s3_endpoint
        if s3_buckets:
            self.s3_buckets = s3_buckets
        self.ceph_admin_key_id = ceph_admin_key_id
        self.ceph_admin_secret_key = ceph_admin_secret_key

        super().init_dsi_config(**kwargs)


class BoxStorageGatewayV53(_StorageGatewayV53):
    model_class = models.BoxStorageGateway

    def init_dsi_config(
        self, boxAppSettings: Dict[str, Any] = None, enterpriseID: str = None, **kwargs
    ):
        self.boxAppSettings = boxAppSettings
        self.enterpriseID = enterpriseID

        super().init_dsi_config(**kwargs)


class BlackPearlStorageGatewayV53(_StorageGatewayV53):
    model_class = models.BlackPearlStorageGateway

    def init_dsi_config(
        self, s3_endpoint: str = None, bp_access_id_file: str = None, **kwargs
    ):
        self.s3_endpoint = s3_endpoint
        self.bp_access_id_file = bp_access_id_file

        super().init_dsi_config(**kwargs)


def StorageGatewayV53(**kwargs) -> models.StorageGateway:
    """
    Pseudo-constructor for the *StorageGatewayV53 classes, which uses the
    connector id from the storage gateway definition document to select which
    type of object to create.
    """

    connector_map = {
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": BlackPearlStorageGatewayV53,
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": BoxStorageGatewayV53,
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": CephStorageGatewayV53,
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": GoogleCloudStorageGatewayV53,
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": GoogleDriveStorageGatewayV53,
        "145812c8-decc-41f1-83cf-bb2a85a2a70b": PosixStorageGatewayV53,
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": S3StorageGatewayV53,
    }

    connector_id = kwargs["connector_id"]
    assert connector_id in connector_map

    sg = connector_map[connector_id](**kwargs)

    return sg.model_class(
        appsync_data={
            "id": sg.id,
            "configuration_hash": storage_gateway_private_schema.dump(sg),
        }
    )
